import unittest

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tensorflow.python.client import device_lib


def np_create_antenna_jones(bsqrt, complex_phase, feed_rotation, ejones):
    """ Compute antenna jones term using numpy """
    result = bsqrt[:,:,None,:,:] * complex_phase[:,:,:,:,None]

    # Reshape npol dimensions to 2x2
    fr_shape = feed_rotation.shape[0:-1] + (2, 2)
    res_shape = result.shape[0:-1] + (2, 2)
    ej_shape = ejones.shape[0:-1] + (2, 2)

    # Multiple result into feed rotation
    # time, antenna, i, j
    # src, time, antenna, channel, j, k
    result = np.einsum("taij,stacjk->stacik",
                       feed_rotation.reshape(fr_shape),
                       result.reshape(res_shape))

    # Multiply result into ejones
    result = np.einsum("stacij,stacjk->stacik",
                       ejones.reshape(ej_shape),result)

    # Return shape in expected format
    return result.reshape(ejones.shape)


class TestCreateAntennaJones(unittest.TestCase):
    """ Tests the CreateAntennaJones operator """

    def setUp(self):
        # Load the rime operation library
        from montblanc.impl.rime.tensorflow import load_tf_lib
        self.rime = load_tf_lib()
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                                if d.device_type == 'GPU']

    def test_create_antenna_jones_operator(self):
        """ Tests the CreateAntennaJones operator """
        # List of type constraint for testing this operator
        type_permutations = [[np.float32, np.complex64],
                             [np.float64, np.complex128]]

        # Run test with the type combinations above
        for FT, CT in type_permutations:
            self._impl_test_create_antenna_jones(FT, CT)

    def _impl_test_create_antenna_jones(self, FT, CT):
        """ Implementation of the CreateAntennaJones operator test """
        rf = lambda *s: np.random.random(size=s).astype(FT)
        rc = lambda *s: (rf(*s) + rf(*s) * 1j).astype(CT)

        nsrc, ntime, na, nchan, npol = 10, 20, 7, 16, 4

        bsqrt = rc(nsrc, ntime, nchan, npol)
        complex_phase = rc(nsrc, ntime, na, nchan)
        feed_rotation = rc(ntime, na, npol)
        ejones = rc(nsrc, ntime, na, nchan, npol)

        np_args = [bsqrt, complex_phase,
                   feed_rotation, ejones]
        arg_names = ["bsqrt", "complex_phase",
                     "feed_rotation", "ejones"]

        tf_args = [tf.Variable(v, name=n) for v, n
                    in zip(np_args, arg_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return self.rime.create_antenna_jones(*tf_args, FT=FT)

        # Pin operation to CPU
        cpu_op = _pin_op('/cpu:0', *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.compat.v1.global_variables_initializer()

        with tf.compat.v1.Session() as S:
            S.run(init_op)

            # Get the CPU sincos
            cpu_aj = S.run(cpu_op)
            np_aj = np_create_antenna_jones(bsqrt,
                complex_phase, feed_rotation, ejones)

            self.assertTrue(np.allclose(np_aj, cpu_aj))

            # Compare with GPU sincos
            for gpu_aj in S.run(gpu_ops):
                self.assertTrue(np.allclose(cpu_aj, gpu_aj))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import functools
import sys
import types
import unittest

import montblanc
import montblanc.util as mbu

from .source_provider import SourceProvider

class NumpySourceProvider(SourceProvider):
    """
    Given a dictionary containing numpy arrays and keyed on array name,
    provides source functions for each array.


    >>> source = NumpySourceProvider({
            "uvw" : np.zeros(shape=(100,14,3),dtype=np.float64),
            "antenna1" : np.zeros(shape=(100,351), dtype=np.int32)
        }, cube)

    >>> context = SourceContext(...)
    >>> source.uvw(context)
    >>> source.antenna1(context)

    """
    def __init__(self, arrays):
        self._arrays = arrays

        def _create_source_function(name, array):
            def _source(self, context):
                """ Generic source function """
                return array[context.array_slice_index(name)]

            return _source

        # Create source methods for each supplied array
        for n, a in list(arrays.items()):
            # Create the source function, update the wrapper,
            # bind it to a method and set the attribute on the object
            f = functools.update_wrapper(
                _create_source_function(n, a),
                _create_source_function)

            f.__doc__ = "Feed function for array '{n}'".format(n=n)

            method = types.MethodType(f, self)
            setattr(self, n, method)

    @property
    def arrays(self):
        return self._arrays

    def __str__(self):
        return self.__class__.__name__

class TestNumpySourceProvider(unittest.TestCase):
    def test_numpy_source_provider(self):
        import hypercube
        import numpy as np

        # Hypercube with ntime, na dimensions
        # and a uvw array
        cube = hypercube.HyperCube()
        cube.register_dimension('ntime', 100)
        cube.register_dimension('na', 64)

        cube.register_array('uvw', ('ntime', 'na', 3), np.float64)

        # Set time and antenna extents
        lt, ut = 10, 50
        la, ua = 10, 20

        # Update dimension extents
        cube.update_dimension('ntime', lower_extent=lt, upper_extent=ut)
        cube.update_dimension('na', lower_extent=la, upper_extent=ua)

        uvw_schema = cube.array('uvw')
        global_uvw_shape = cube.dim_global_size(*uvw_schema.shape)
        uvw = (np.arange(np.product(global_uvw_shape))
                    .reshape(global_uvw_shape)
                    .astype(np.float64))

        # Create a Numpy Source Provider
        source_prov = NumpySourceProvider({"uvw" : uvw})

        class Context(object):
            """ Mock up a context object """
            def __init__(self, array, cube):
                self._cube = cube
                self.array_slice_index = cube.array_slice_index

                array_schema = cube.array(array)

                self.shape = cube.dim_extent_size(*array_schema.shape)
                self.dtype = array_schema.dtype


        data = source_prov.uvw(Context('uvw', cube))
        uvw_slice = uvw[lt:ut, la:ua, :]

        # Check that we've got the shape defined by
        # cube extents and the given dtype
        self.assertTrue(np.all(data == uvw_slice))
        self.assertTrue(data.shape == uvw_slice.shape)
        self.assertTrue(data.dtype == uvw_slice.dtype)

if __name__ == "__main__":
    unittest.main()



import collections
import itertools
import sys
import six

import attridict as AttrDict

import tensorflow as tf

def _get_queue_types(fed_arrays, data_sources):
    """
    Given a list of arrays to feed in fed_arrays, return
    a list of associated queue types, obtained from tuples
    in the data_sources dictionary
    """
    try:
        return [data_sources[n].dtype for n in fed_arrays]
    except KeyError as e:
        six.reraise(ValueError, ValueError("Array '{k}' has no data source!"), sys.exc_info()[2])

class QueueWrapper(object):
    def __init__(self, name, queue_size, fed_arrays, data_sources, shared_name=None):
        self._name = name
        self._fed_arrays = fed_arrays
        self._data_sources = data_sources

        # Infer types of the given fed_arrays
        self._queue_types = _get_queue_types(fed_arrays, data_sources)

        # Create placeholders for the fed arrays
        self._placeholders = [tf.compat.v1.placeholder(dt, name="{n}_placeholder".format(n=n))
            for n, dt in zip(fed_arrays, self._queue_types)]

        # Create a FIFOQueue of a given size with the supplied queue types
        self._queue = tf.queue.FIFOQueue(queue_size,
            self._queue_types, name=name, shared_name=shared_name)

        # Create enqueue operation using placeholders
        self._enqueue_op = self._queue.enqueue(self.placeholders)

        self._dequeue_op = self._queue.dequeue()

        self._close_op = self._queue.close()

        self._size_op = self._queue.size()

    @property
    def name(self):
        return self._name

    @property
    def queue_types(self):
        return self._queue_types

    @property
    def placeholders(self):
        return self._placeholders

    @property
    def queue(self):
        return self._queue

    @property
    def fed_arrays(self):
        return self._fed_arrays

    @property
    def enqueue_op(self):
        return self._enqueue_op

    @property
    def dequeue_op(self):
        return self._dequeue_op

    def dequeue(self):
        return self._queue.dequeue()

    def dequeue_to_dict(self):
        return {k: v for k, v in zip(
            self._fed_arrays, self._queue.dequeue())}

    def dequeue_to_attrdict(self):
        return AttrDict((k, v) for k, v in zip(
            self._fed_arrays, self._queue.dequeue()))

    @property
    def close_op(self):
        return self._close_op

    def close(self):
        return self._queue.close()

    @property
    def size_op(self):
        return self._size_op

    def size(self):
        return self._queue.size()

    def __str__(self):
        return 'Queue of size {s} with with types {t}'.format(
            s=len(self._queue_types),
            t=self._queue_types)

class SingleInputMultiQueueWrapper(QueueWrapper):
    def __init__(self, name, queue_size, fed_arrays, data_sources,
        shared_name=None, count=4):

        super(SingleInputMultiQueueWrapper, self).__init__(name, queue_size,
            fed_arrays, data_sources, shared_name)

        R = list(range(1, count))

        extra_names = ['%s_%s' % (name, i) for i in R]
        extra_shared_names = ['%s_%s' % (shared_name, i) for i in R]

        extra_queues = [tf.queue.FIFOQueue(queue_size, self._queue_types,
                name=n, shared_name=sn)
            for n, sn in zip(extra_names, extra_shared_names)]

        extra_enqueue_ops = [q.enqueue(self._placeholders) for q in extra_queues]
        extra_dequeue_ops = [q.dequeue() for q in extra_queues]
        extra_close_ops = [q.close() for q in extra_queues]
        extra_size_ops = [q.size() for q in extra_queues]

        self._names = [self._name] + extra_names
        self._queues = [self._queue] + extra_queues
        self._enqueue_ops = [self._enqueue_op] + extra_enqueue_ops
        self._dequeue_ops = [self._dequeue_op] + extra_dequeue_ops
        self._size_ops = [self._size_op] + extra_size_ops

    @property
    def name(self):
        return self._names

    @property
    def queue(self):
        return self._queues

    @property
    def enqueue_op(self):
        return self._enqueue_ops

    @property
    def dequeue_op(self):
        return self._dequeue_ops

    @property
    def close_op(self):
        return self._close_ops


def create_queue_wrapper(name, queue_size, fed_arrays, data_sources, *args, **kwargs):
    """
    Arguments
        name: string
            Name of the queue
        queue_size: integer
            Size of the queue
        fed_arrays: list
            array names that will be fed by this queue
        data_sources: dict
            (lambda/method, dtype) tuples, keyed on array names

    """

    qtype = SingleInputMultiQueueWrapper if 'count' in kwargs else QueueWrapper
    return qtype(name, queue_size, fed_arrays, data_sources, *args, **kwargs)




import argparse
import json
import logging
import os
import socket

CHUNK_SIZE = 1024
DEFAULT_PORT = 8888
PING = 'PING'
PONG = 'PONG'

# Set up logging
logging.basicConfig(level=logging.DEBUG, format='%(levelname)s - %(message)s')

# Command line parser
parser = argparse.ArgumentParser()
parser.add_argument('server', type=str)
parser.add_argument('-c', '--clean', default=True, type=bool)
parser.add_argument('--no-start', dest='start', action='store_false')
parser.add_argument('--start', dest='start', action='store_true')
parser.set_defaults(start=False)
args = parser.parse_args()

port_index = args.server.rfind(':')

if port_index != -1:
    port = int(args.server[port_index+1:])
    address = (args.server[:port_index], port)
else:
    address = (args.server, DEFAULT_PORT)

chunks = []

try:
    logging.info("Connecting to {a}".format(a=address))
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(address)

    # Wait for the server to signal specification is being sent
    ping = s.recv(len(PING))

    if ping != PING:
        raise ValueError("Expected {p}, received {i}".format(p=PING, i=ping))

    # Signal server we're ready for data
    s.send(PONG)

    logging.info("Receiving for cluster specification")

    while True:
        data = s.recv(CHUNK_SIZE)

        if not data:
            break

        chunks.append(data)

except:
    logging.error("Exception")
    raise

finally:
    logging.info('Closing {a}'.format(a=address))
    s.shutdown(socket.SHUT_RDWR)
    s.close()

try:
    # Join chunks into string, parse json
    # and extract cluster specification, job and task
    data = json.loads(''.join(chunks))
    cluster, job, task = (data[v] for v in ('cluster', 'job', 'task'))
except KeyError as e:
    logging.error("Key '{k}' not in dictionary".format(k=e.message))
    raise

if args.start is True:
    import tensorflow as tf
    import numpy as np
    import time

    server = tf.distribute.Server(cluster, job_name=job, task_index=task)
    logging.info("Server Target is '{st}'".format(st=server.target))

    g = tf.Graph()

    with g.as_default():
        with tf.compat.v1.container('shared'):
            queue_in = tf.queue.FIFOQueue(10, [tf.int32],
                name='queue_in',
                shared_name='master_queue_in')

            queue_out = tf.queue.FIFOQueue(10, [tf.string],
                name='queue_out',
                shared_name='master_queue_out')

            tmp = tf.Variable(-1, tf.float32, name='master_tmp')

        do_deq = queue_in.dequeue()
        do_enq = queue_out.enqueue("Hello World")

    with tf.compat.v1.Session(server.target, graph=g) as S:
        S.run(tf.compat.v1.initialize_local_variables())
        print((S.run([do_deq])))
        print((S.run([do_deq])))
        print((S.run([do_deq])))
        print((S.run([do_deq])))

        print(('Value of master_tmp={mt}.'.format(mt=S.run(tmp))))

        S.run(do_enq)

        time.sleep(2)





import unittest

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tensorflow.python.client import device_lib


class TestFeedRotation(unittest.TestCase):
    """ Tests the FeedRotation operator """

    def setUp(self):
        # Load the rime operation library
        from montblanc.impl.rime.tensorflow import load_tf_lib
        self.rime = load_tf_lib()
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                         if d.device_type == 'GPU']

    def test_feed_rotation(self):
        """ Test the FeedRotation operator """
        # List of type constraint for testing this operator
        type_permutations = [
            [np.float32, np.complex64, 'linear'],
            [np.float64, np.complex128, 'linear'],
            [np.float32, np.complex64, 'circular'],
            [np.float64, np.complex128, 'circular'],
        ]

        # Run test with the type combinations above
        for FT, CT, feed_type in type_permutations:
            self._impl_test_feed_rotation(FT, CT, feed_type)

    def _impl_test_feed_rotation(self, FT, CT, feed_type):
        """ Implementation of the FeedRotation operator test """

        # Create input variables
        ntime, na = 10, 7

        parallactic_angle = np.random.random(size=[ntime, na]).astype(FT)
        parallactic_angle_sin = np.sin(parallactic_angle)
        parallactic_angle_cos = np.cos(parallactic_angle)

        # Argument list
        np_args = [parallactic_angle_sin, parallactic_angle_cos]
        # Argument string name list
        arg_names = ['parallactic_angle_sin', 'parallactic_angle_cos']
        # Constructor tensorflow variables
        tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, arg_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return self.rime.feed_rotation(*tf_args, CT=CT,
                                               feed_type=feed_type)

        # Pin operation to CPU
        cpu_op = _pin_op('/cpu:0', *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.compat.v1.global_variables_initializer()

        with tf.compat.v1.Session() as S:
            S.run(init_op)

            cpu_feed_rotation = S.run(cpu_op)

            for gpu_feed_rotation in S.run(gpu_ops):
                self.assertTrue(np.allclose(cpu_feed_rotation,
                                            gpu_feed_rotation))


if __name__ == "__main__":
    unittest.main()



"""
This script runs Montblanc and MeqTrees and compares
the visibilities output by both.

To run this script you'll need a Measurement Set.
It's often easier to create this using the
`simms <https://github.com/radio-astro/simms_>`_ package.

For example, create a VLA Measurement Set, call the following:

.. code-block:: shell

    simms -T vla -t ascii -cs itrf -st 1 -dt 4 -f0 1.42GHz -df 4MHz
        -nc 8 -dir "J2000,5h42m36.1378s,+49d51m7.23000000001s"
        -feed 'perfect R L' -pl "RR RL LR LL" -n vla_test.ms
        ~/paper_sims/simms/simms/observatories/vlac.itrf.txt

Then, to call the script run

.. code-block:: shell

    python test_meq_tf.py vla.ms -p circular -a "overwrite_beams=True"

This tells the script to configure Montblanc+MeqTrees for circular
polarisations (the VLA telescope uses circular polarisations) and
to create new cos**3 testing beams. It is possible to also supply
your own beams for the test by specifying
``-a "beam_file_schema='my_beam_\$(corr)_\$(reim).fits'`` for example.

Other options can be passed

-- See :func:`run_test` for more details
and options.
"""


import itertools
import os
import subprocess
import sys
import tempfile

from astropy.io import fits
import numpy as np
import pyrap.tables as pt

rf = np.random.random

from montblanc.tests.beam_factory import beam_factory

# Directory that holds MS and Beam data
DATA_DIR = 'data'

def run_test(msfile="/mb_testing/data/mk64.Lwide.0.5hr.30s.856mhz.ms", 
             pol_type="linear", **kwargs):
    """
    Parameters
    ----------
    msfile : str
        Name of the Measurement Set
    pol_type : str
        'linear' or 'circular'
    beam_file_schema (optional) : str
        Beam filename schema. Defaults to 'test_beam_$(corr)_$(reim).fits'
    overwrite_beams (optional) : bool
        If ``True`` create new beams using the cos**3 beam
    """


    #=========================================
    # Directory and Script Configuration
    #=========================================

    # Directory in which we expect our measurement set to be located
    meq_vis_column = 'MODEL_DATA'
    mb_vis_column = 'CORRECTED_DATA'

    # Directory in which meqtree-related files are read/written
    import os
    meq_dir = os.path.join(os.path.dirname(__file__), 'meqtrees')
    # Scripts
    meqpipe = 'meqtree-pipeliner.py'
    # Meqtree profile and script
    cfg_file = os.path.join(meq_dir, 'tdlconf.profiles')
    sim_script = os.path.join(meq_dir, 'turbo-sim.py')
    tigger_sky_file = os.path.join(meq_dir, 'sky_model.txt')

    # Is the beam enabled
    beam_on = kwargs.get('beam_on', True)
    beam_on = 1 if beam_on is True else 0

    # Directory in which we expect our beams to be located
    beam_file_schema = 'test_beam_$(corr)_$(reim).fits'

    # Beam file pattern
    beam_file_schema = kwargs.get("beam_file_schema", beam_file_schema)

    l_axis = kwargs.get('l_axis', '-X')
    m_axis = kwargs.get('m_axis', 'Y')

    # Find the location of the meqtree pipeliner script
    meqpipe_actual = subprocess.check_output(['which', meqpipe]).strip()
    cfg_section = '-'.join(('montblanc', 'compare', pol_type))

    #======================================================
    # Configure the beam files with frequencies from the MS
    #======================================================

    from montblanc.impl.rime.tensorflow.sources.fits_beam_source_provider import (
        _create_filenames, _open_fits_files)

    # Zero the visibility data
    with pt.table(msfile, ack=False, readonly=False) as T:
        data_desc = T.getcoldesc('DATA')

        try:
            shape = data_desc['shape'].tolist()
        except KeyError:
            shape = list(T.getcol('DATA', startrow=0, nrow=1).shape[1:])

        shape = [T.nrows()] + shape
        T.putcol(mb_vis_column, np.zeros(shape, dtype=np.complex64))
        T.putcol(meq_vis_column, np.zeros(shape, dtype=np.complex64))

    # Extract frequencies from the MS
    with pt.table(msfile + '::SPECTRAL_WINDOW', ack=False) as SW:
        frequency = SW.getcol('CHAN_FREQ')[0]

    bandwidth = frequency[-1] - frequency[0]

    overwrite_beams = kwargs.get('overwrite_beams', False)

    # Get filenames from pattern and open the files
    filenames = beam_factory(polarisation_type=pol_type,
                            frequency=frequency,
                            schema=beam_file_schema,
                            overwrite=overwrite_beams)

    #=========================================
    # Source Configuration
    #=========================================

    np.random.seed(0)
    dtype = np.float64
    ctype = np.complex128 if dtype == np.float64 else np.complex64

    def get_point_sources(nsrc):
        source_coords = np.empty(shape=(nsrc, 2), dtype=dtype)
        stokes = np.empty(shape=(nsrc, 4), dtype=dtype)
        I, Q, U, V = stokes[:,0], stokes[:,1], stokes[:,2], stokes[:,3]
        alphas = np.empty(shape=(nsrc,), dtype=dtype)
        ref_freq = np.empty(shape=(nsrc,), dtype=dtype)

        # Source coordinates between -45 and 45 degrees
        source_coords[:] = (rf(size=source_coords.shape) - 0.5)*90.0
        Q[:] = rf(size=Q.shape)*0.1
        U[:] = rf(size=U.shape)*0.1
        V[:] = rf(size=V.shape)*0.1
        I[:] = np.sqrt(Q**2 + U**2 + V**2)*1.5 + rf(size=I.shape)*0.1

        # Zero and invert selected stokes parameters
        if nsrc > 0:
            zero_srcs = np.random.randint(nsrc, size=(2,))
            source_coords[zero_srcs,:] = 0

            # Create sources with both positive and negative flux
            sign = 2*np.random.randint(2, size=I.shape) - 1
            I[:] *= sign

        alphas[:] = 2*(np.random.random(size=alphas.size) - 0.5)

        ref_freq[:] = 1.3e9 + np.random.random(ref_freq.size)*0.2e9

        return (np.deg2rad(source_coords), np.asarray(stokes),
                np.asarray(alphas), np.asarray(ref_freq))

    def get_gaussian_sources(nsrc):
        c, s, a, r= get_point_sources(nsrc)
        gauss_shape = np.empty(shape=(3, nsrc), dtype=np.float64)
        gauss_shape[:] = rf(size=gauss_shape.shape)
        return c, s, a, r, gauss_shape

    npsrc, ngsrc = 10, 10

    pt_lm, pt_stokes, pt_alpha, pt_ref_freq = get_point_sources(npsrc)

    assert pt_lm.shape == (npsrc, 2), pt_lm.shape
    assert pt_stokes.shape == (npsrc, 4), pt_stokes.shape
    assert pt_alpha.shape == (npsrc,), pt_alpha.shape
    assert pt_ref_freq.shape == (npsrc,), pt_ref_freq.shape

    g_lm, g_stokes, g_alpha, g_ref_freq, g_shape = get_gaussian_sources(ngsrc)

    #=========================================
    # Create Tigger ASCII sky model
    #=========================================

    from Tigger.Models.Formats.AIPSCCFITS import lm_to_radec

    # Need the phase centre for lm_to_radec
    with pt.table(msfile + '::FIELD', ack=False, readonly=True) as F:
        ra0, dec0 = F.getcol('PHASE_DIR')[0][0]

    # Create the tigger sky model
    with open(tigger_sky_file, 'w') as f:
        f.write('#format: ra_d dec_d i q u v spi freq0 emaj_s emin_s pa_d\n')
        it = enumerate(zip(pt_lm, pt_stokes, pt_alpha, pt_ref_freq))
        for i, ((l, m), (I, Q, U, V), alpha, ref_freq) in it:
            ra, dec = lm_to_radec(l, m, ra0, dec0)
            l, m = np.rad2deg([ra,dec])

            f.write('{l:.20f} {m:.20f} {i} {q} {u} {v} {spi} {rf:.20f}\n'.format(
                l=l, m=m, i=I, q=Q, u=U, v=V, spi=alpha, rf=ref_freq))

        it = enumerate(zip(g_lm, g_stokes, g_alpha, g_ref_freq, g_shape.T))
        for i, ((l, m), (I, Q, U, V), alpha, ref_freq, (emaj, emin, pa)) in it:
            ra, dec = lm_to_radec(l, m, ra0, dec0)
            l, m = np.rad2deg([ra,dec])
            # Convert to seconds
            emaj, emin = np.asarray([emaj, emin])*648000./np.pi
            # Convert to degrees
            pa *= 180.0/np.pi

            f.write('{l:.20f} {m:.20f} {i} {q} {u} {v} {spi} {rf:.20f} '
                    '{emaj} {emin} {pa}\n'.format(
                        l=l, m=m, i=I, q=Q, u=U, v=V, spi=alpha, rf=ref_freq,
                        emaj=emaj, emin=emin, pa=pa))

#=========================================
# Call MeqTrees
#=========================================

    #=========================================
    # Call MeqTrees
    #=========================================

    cmd_list = ['python3',
        # Meqtree Pipeline script
        meqpipe_actual,
        # Configuration File
        '-c', cfg_file,
        # Configuration section
        '[{section}]'.format(section=cfg_section),
        # Enable the beam?
        'me.e_enable = {e}'.format(e=beam_on),
        # Measurement Set
        'ms_sel.msname={ms}'.format(ms=msfile),
        # Tigger sky file
        'tiggerlsm.filename={sm}'.format(sm=tigger_sky_file),
        # Output column
        'ms_sel.output_column={c}'.format(c=meq_vis_column),
        # Imaging Column
        'img_sel.imaging_column={c}'.format(c=meq_vis_column),
        # Beam FITS file pattern
        'pybeams_fits.filename_pattern={p}'.format(p=beam_file_schema),
        # FITS L and M AXIS
        'pybeams_fits.l_axis={l}'.format(l=l_axis),
        'pybeams_fits.m_axis={m}'.format(m=m_axis),
        sim_script,
        '=simulate'
        ]

    import montblanc

    from montblanc.impl.rime.tensorflow.ms import MeasurementSetManager
    from montblanc.impl.rime.tensorflow.sources import (SourceProvider,
        MSSourceProvider,
        FitsBeamSourceProvider,
        CachedSourceProvider)

    from montblanc.impl.rime.tensorflow.sinks import MSSinkProvider

    class RadioSourceProvider(SourceProvider):
        def name(self):
            return "RadioSourceProvider"

        def point_lm(self, context):
            lp, up = context.dim_extents('npsrc')
            return pt_lm[lp:up, :]

        def point_stokes(self, context):
            (lp, up), (lt, ut), (lc, uc) = context.dim_extents('npsrc', 'ntime', 'nchan')
            # (npsrc, ntime, nchan, 4)
            s = pt_stokes[lp:up,None,None,:]
            a = np.broadcast_to(pt_alpha[lp:up,None,None,None], (up-lp,ut-lt,1,1))
            rf = pt_ref_freq[lp:up,None,None,None]
            f = frequency[None,None,lc:uc,None]

            return s*(f/rf)**a

        def gaussian_lm(self, context):
            lg, ug = context.dim_extents('ngsrc')
            return g_lm[lg:ug, :]

        def gaussian_stokes(self, context):
            (lg, ug), (lt, ut), (lc, uc) = context.dim_extents('ngsrc', 'ntime', 'nchan')
            # (ngsrc, ntime, nchan, 4)
            s = g_stokes[lg:ug,None,None,:]
            a = np.broadcast_to(pt_alpha[lg:ug,None,None,None], (ug-lg,ut-lt,1,1))
            rf = g_ref_freq[lg:ug,None,None,None]
            f = frequency[None,None,lc:uc,None]

            return s*(f/rf)**a

        def gaussian_shape(self, context):
            (lg, ug) = context.dim_extents('ngsrc')
            gauss_shape = g_shape[:,lg:ug]
            emaj = gauss_shape[0]
            emin = gauss_shape[1]
            pa = gauss_shape[2]

            gauss = np.empty(context.shape, dtype=context.dtype)

            gauss[0,:] = emaj * np.sin(pa)
            gauss[1,:] = emaj * np.cos(pa)
            emaj[emaj == 0.0] = 1.0
            gauss[2,:] = emin / emaj

            return gauss

        def updated_dimensions(self):
            return [('npsrc', pt_lm.shape[0]), ('ngsrc', g_lm.shape[0])]

    slvr_cfg = montblanc.rime_solver_cfg(
        mem_budget=1024*1024*1024,
        data_source='default',
        dtype='double' if dtype == np.float64 else 'float',
        polarisation_type=pol_type,
        auto_correlations=False,
        version='tf')

    slvr = montblanc.rime_solver(slvr_cfg)

    ms_mgr = MeasurementSetManager(msfile, slvr_cfg)

    source_providers = []
    source_providers.append(MSSourceProvider(ms_mgr))

    if beam_on == 1:
        beam_prov = FitsBeamSourceProvider(beam_file_schema,
            l_axis=l_axis, m_axis=m_axis)
        source_providers.append(beam_prov)

    source_providers.append(RadioSourceProvider())
    cache_prov = CachedSourceProvider(source_providers)
    source_providers = [cache_prov]

    sink_providers = [MSSinkProvider(ms_mgr, mb_vis_column)]
    slvr.solve(source_providers=source_providers,
        sink_providers=sink_providers)

    import time
    time.sleep(1)

    for obj in source_providers + sink_providers + [ms_mgr]:
        obj.close()

    # Call the meqtrees simulation script, dumping visibilities into MODEL_DATA
    subprocess.call(cmd_list)

    # Compare MeqTree and Montblanc visibilities
    with pt.table(msfile, ack=False, readonly=True) as MS:
        ntime, nbl, nchan = slvr.hypercube.dim_global_size('ntime', 'nbl', 'nchan')
        shape = (ntime, nbl, nchan, 4)
        meq_vis = MS.getcol(meq_vis_column).reshape(shape)
        mb_vis = MS.getcol(mb_vis_column).reshape(shape)

        # Compare
        close = np.isclose(meq_vis, mb_vis)
        not_close = np.invert(close)
        problems = np.nonzero(not_close)

        # Everything agrees, exit
        if problems[0].size == 0:
            print('Montblanc and MeqTree visibilities agree')
            sys.exit(1)

        bad_vis_file = 'bad_visibilities.txt'

        # Some visibilities differ, do some analysis
        print((("Montblanc differs from MeqTrees by {nc}/{t} visibilities. "
            "Writing them out to '{bvf}'").format(
            nc=problems[0].size, t=not_close.size, bvf=bad_vis_file)))

        abs_diff = np.abs(meq_vis - mb_vis)
        rmsd = np.sqrt(np.sum(abs_diff**2)/abs_diff.size)
        nrmsd = rmsd / (np.max(abs_diff) - np.min(abs_diff))
        print(('RMSD {rmsd} NRMSD {nrmsd}'.format(rmsd=rmsd, nrmsd=nrmsd)))

        # Plot a histogram of the difference
        try:
            import matplotlib
            matplotlib.use('pdf')
            import matplotlib.pyplot as plt
        except:
            print(("Exception importing matplotlib %s" % sys.exc_info()[2]))
        else:
            try:
                nr_of_bins = 100
                n, bins, patches = plt.hist(abs_diff.flatten(),
                    bins=np.logspace(np.log10(1e-10), np.log10(1.0), nr_of_bins))

                plt.gca().set_xscale("log")
                plt.xlabel('Magnitude Difference')
                plt.ylabel('Counts')
                plt.grid(True)

                plt.savefig('histogram.pdf')
            except:
                print(("Error plotting histogram %s" % sys.exc_info()[2]))

        mb_problems = mb_vis[problems]
        meq_problems = meq_vis[problems]
        difference = mb_problems - meq_problems
        amplitude = np.abs(difference)

        # Create an iterator over the first 100 problematic visibilities
        t = (np.asarray(problems).T, mb_problems, meq_problems, difference, amplitude)
        it = enumerate(zip(*t))
        it = itertools.islice(it, 0, 1000, 1)

        # Write out the problematic visibilities to file
        with open(bad_vis_file, 'w') as f:
            for i, (p, mb, meq, d, amp) in it:
                f.write("{i} {t} Montblanc: {mb} MeqTrees: {meq} "
                    "Difference {d} Absolute Difference {ad} \n".format(
                        i=i, t=p, mb=mb, meq=meq, d=d, ad=amp))


if __name__ == "__main__":
    import argparse
    from os.path import join as pjoin

    from montblanc.util import parse_python_assigns

    def create_parser():
        p = argparse.ArgumentParser()
        p.add_argument("ms", default=pjoin("data", "mk64.Lwide.0.5hr.30s.856mhz.ms"),
                                nargs="?")
        p.add_argument("-p", "--polarisation-type",
                            choices=['linear', 'circular'],
                            default='linear')
        p.add_argument("-a", "--args", type=parse_python_assigns,
                            default="",
                            help="semi-colon separated list of "
                                "python variable assignments. "
                                "These variable assignments are "
                                "passed into the run_tests function."
                                "See its docstring for details.")
        return p

    args = create_parser().parse_args()

    run_test(args.ms, args.polarisation_type, **args.args)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import textwrap

_desc_wrapper = textwrap.TextWrapper(initial_indent=" "*4,
    subsequent_indent=" "*8, width=70)

_help_wrapper = textwrap.TextWrapper(width=70)

def context_help(context, display_cube=False):
    from montblanc.impl.rime.tensorflow.sources.source_context import SourceContext
    from montblanc.impl.rime.tensorflow.sinks.sink_context import SinkContext

    if isinstance(context, SourceContext):
        ctx_type = 'source'
        behaviour1 = 'requesting'
        behaviour2 = 'return'
        shape = context.shape
        dtype = context.dtype
    elif isinstance(context, SinkContext):
        ctx_type = 'sink'
        behaviour1 = 'providing'
        behaviour2 = 'with'
        shape = context.data.shape
        dtype = context.data.dtype
    else:
        raise TypeError("Invalid context {t}".format(t=type(context)))

    description = context._array_schema.get('description', 'No Description')
    units = context._array_schema.get('units', 'None Specified')
    schema = context._array_schema.shape
    cube = context._cube
    cube_dims = cube.dimensions(copy=False)
    global_shape = tuple([cube.dim_global_size(d) if d in cube_dims
        else d for d in schema])
    l_extents = tuple([cube.dim_lower_extent(d) if d in cube_dims
        else 0 for d in schema])
    u_extents = tuple([cube.dim_upper_extent(d) if d in cube_dims
        else d for d in schema])

    dim_pad = " "*12
    wrap = _desc_wrapper.wrap

    lines = []
    lines.append("'{name}' data source information:".format(
        name=context._name))
    lines += wrap("Description: {description}".format(
        description=description))
    lines += wrap("Units: {units}".format(units=units))
    lines += wrap("Schema or abstract shape: {schema}\n".format(schema=schema))
    lines += ["{p}where '{s}' is '{d}'".format(
            p=dim_pad, s=d, d=cube_dims[d].description)
        for d in schema if d in cube_dims]
    lines += wrap("Global shape on this iteration: "
        "{global_shape}\n".format(global_shape=global_shape))
    lines += wrap("Local shape for this context: "
        "{local_shape}\n".format(local_shape=shape))
    lines += wrap("Lower extents within global shape: "
        "{lower_extents}\n".format(lower_extents=l_extents))
    lines += wrap("Upper extents within global shape: "
        "{upper_extents}\n".format(upper_extents=u_extents))
    lines.append('\n')

    dims, strides = list(zip(*context._iter_args))

    lines.append("Iteration information:")
    lines += wrap("Iterating over the {d} "
        "dimensions with global sizes of {gs} "
        "in strides of {s}.".format(
            d=dims, s=strides, gs=tuple(cube.dim_global_size(*dims))))
    lines.append('\n')

    wrap = _help_wrapper.wrap

    lines += wrap("This context is {b1} the '{name}' data {ctype} "
        "{b2} an array of shape '{shape}' and dtype '{dtype}'. "
        "This portion of data lies between the lower '{lex}' "
        " and upper '{uex}' extent of a global shape of '{gs}'. "
        "The abstract shape of this data source is {schema}.".format(
            name=context._name, ctype=ctx_type,
            b1=behaviour1, b2=behaviour2,
            shape=shape, dtype=dtype,
            schema=schema, lex=l_extents, uex=u_extents, gs=global_shape))

    if display_cube is True:
        lines += ['', 'Hypercube', '', str(cube)]

    return '\n'.join(lines)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import collections
import copy
import itertools
from os import lockf
import threading
import sys
import types
import six

import concurrent.futures as cf
import numpy as np
import tensorflow as tf
from tensorflow.python.client import timeline

import attridict as AttrDict

import attr

import montblanc
import montblanc.util as mbu
from montblanc.src_types import source_var_types
from montblanc.solvers import MontblancTensorflowSolver

from . import load_tf_lib
from .cube_dim_transcoder import CubeDimensionTranscoder
from .staging_area_wrapper import create_staging_area_wrapper
from .sources import (SourceContext, DefaultsSourceProvider)
from .sinks import (SinkContext, NullSinkProvider)
from .start_context import StartContext
from .stop_context import StopContext
from .init_context import InitialisationContext

ONE_KB, ONE_MB, ONE_GB = 1024, 1024**2, 1024**3

QUEUE_SIZE = 10

rime = load_tf_lib()

DataSource = attr.make_class("DataSource", ['source', 'dtype', 'name'],
    slots=True, frozen=True)
DataSink = attr.make_class("DataSink", ['sink', 'name'],
    slots=True, frozen=True)
FeedOnce = attr.make_class("FeedOnce", ['ph', 'var', 'assign_op'],
    slots=True, frozen=True)

class RimeSolver(MontblancTensorflowSolver):
    """ RIME Solver Implementation """

    def __init__(self, slvr_cfg):
        """
        RimeSolver Constructor

        Parameters:
            slvr_cfg : SolverConfiguration
                Solver Configuration variables
        """
        super(RimeSolver, self).__init__(slvr_cfg)

        #=========================================
        # Register hypercube Dimensions
        #=========================================

        cube, slvr_cfg = self.hypercube, self.config()

        _setup_hypercube(cube, slvr_cfg)

        #=======================
        # Data Sources and Sinks
        #=======================

        # Get the defaults data source (default or test data)
        data_source = slvr_cfg['data_source']
        montblanc.log.info("Defaults Data Source '{}'".format(data_source))

        # Construct list of data sources and sinks
        # internal to the solver.
        # These will be overridden by source and sink
        # providers supplied by the user in the solve()
        # method
        default_prov = _create_defaults_source_provider(cube, data_source)
        self._source_providers = [default_prov]
        self._sink_providers = [NullSinkProvider()]

        #==================
        # Data Source Cache
        #==================

        class SourceCache(object):
            def __init__(self):
                self._cache = {}
                self._lock = threading.Lock()

            def __getitem__(self, key):
                key = hash(bytes(key))
                with self._lock:
                    return self._cache[key]

            def __setitem__(self, key, value):
                key = hash(bytes(key))
                with self._lock:
                    self._cache[key]=value

            def __delitem__(self, key):
                key = hash(bytes(key))
                with self._lock:
                    del self._cache[key]

            def pop(self, key, default=None):
                key = hash(bytes(key))
                with self._lock:
                    return self._cache.pop(key, default)


        self._source_cache = SourceCache()

        #==================
        # Memory Budgeting
        #==================

        # For deciding whether to rebudget
        self._previous_budget = 0
        self._previous_budget_dims = {}

        #================
        # Cube Transcoder
        #================
        self._iter_dims = ['ntime', 'nbl']
        self._transcoder = CubeDimensionTranscoder(self._iter_dims)

        #================================
        # Staging Area Data Source Configuration
        #================================

        dfs = { n: a for n, a in list(cube.arrays().items())
            if not 'temporary' in a.tags }

        # Descriptors are not user-defined arrays
        # but a variable passed through describing a chunk of the
        # problem. Make it look as if it's an array
        if 'descriptor' in dfs:
            raise KeyError("'descriptor' is reserved, "
                "please use another array name.")

        dfs['descriptor'] = AttrDict(dtype=np.int32)

        #=========================
        # Tensorflow devices
        #=========================

        from tensorflow.python.client import device_lib
        devices = device_lib.list_local_devices()

        device_type = slvr_cfg['device_type'].upper()

        gpus = [d.name for d in devices if d.device_type == 'GPU']
        cpus = [d.name for d in devices if d.device_type == 'CPU']

        if device_type == 'GPU' and len(gpus) == 0:
            montblanc.log.warning("No GPUs are present, falling back to CPU.")
            device_type = 'CPU'

        use_cpus = device_type == 'CPU'
        montblanc.log.info("Using '{}' devices for compute".format(device_type))
        self._devices = cpus if use_cpus else gpus
        self._shards_per_device = spd = 2
        self._nr_of_shards = shards = len(self._devices)*spd
        # shard_id == d*spd + shard
        self._shard = lambda d, s: d*spd + s

        assert len(self._devices) > 0

        #=========================
        # Tensorflow Compute Graph
        #=========================

        # Create all tensorflow constructs within the compute graph
        with tf.Graph().as_default() as compute_graph:
            # Create our data feeding structure containing
            # input/output staging_areas and feed once variables
            self._tf_feed_data = _construct_tensorflow_feed_data(
                dfs, cube, self._iter_dims, shards)

            # Construct tensorflow expressions for each shard
            self._tf_expr = [_construct_tensorflow_expression(
                    slvr_cfg,
                    self._tf_feed_data, dev, self._shard(d,s))
                for d, dev in enumerate(self._devices)
                for s in range(self._shards_per_device)]

            # Now forbid modification of the graph
            compute_graph.finalize()

        #==========================================
        # Tensorflow Session
        #==========================================

        # Create the tensorflow session object
        # Use supplied target, if present
        tf_server_target = slvr_cfg.get('tf_server_target', '')

        montblanc.log.debug("Attaching session to tensorflow server "
            "'{tfs}'".format(tfs=tf_server_target))

        session_config = tf.compat.v1.ConfigProto(allow_soft_placement=True)

        self._tf_session = tf.compat.v1.Session(tf_server_target,
            graph=compute_graph, config=session_config)

        #======================
        # Thread pool executors
        #======================
        tpe = cf.ThreadPoolExecutor

        self._descriptor_executor = tpe(1)
        self._feed_executors = [tpe(1) for i in range(shards)]
        self._compute_executors = [tpe(1) for i in range(shards)]
        self._consumer_executor = tpe(1)
        
        class InputsWaiting(object):
            """
            Keep track of the number of inputs waiting
            to be consumed on each shard

            Must be called inside __enter__ of _feederqueuelock
            """
            def __init__(self, shards):
                self._feederqueuelock = [threading.Lock()] * shards
                self._inputs_waiting = np.zeros(shape=(shards,), dtype=np.int32)

            def get(self, shard, skiplock=False):
                if not skiplock:
                    with self._feederqueuelock[shard]:
                        return self._inputs_waiting[shard]
                else:
                    return self._inputs_waiting[shard]

            def getlock(self, shard):
                return self._feederqueuelock[shard]

            def increment(self, shard, skiplock=False):
                if not skiplock:
                    with self._feederqueuelock[shard]:
                        self._inputs_waiting[shard] += 1
                else:
                    self._inputs_waiting[shard] += 1

            def decrement(self, shard, skiplock=False):
                if not skiplock:
                    with self._feederqueuelock[shard]:
                        self._inputs_waiting[shard] -= 1
                else:
                    self._inputs_waiting[shard] -= 1

        self._inputs_waiting = InputsWaiting(shards)

        #======================
        # Tracing
        #======================

        class RunMetaData(object):
            def __init__(self):
                self._rm = []
                self._lock = threading.Lock()

            def clear(self):
                with self._lock:
                    self._rm = []

            def save(self, run_metadata):
                with self._lock:
                    self._rm.append(run_metadata)

            def write(self, tag=None):
                with self._lock:
                    if len(self._rm) == 0:
                        return

                    if tag is None:
                        tag='0'

                    metadata = tf.compat.v1.RunMetadata()
                    [metadata.MergeFrom(m) for m in self._rm]

                    tl = timeline.Timeline(metadata.step_stats)
                    trace_filename = 'compute_timeline_%d.json' % tag
                    with open(trace_filename, 'w') as f:
                        f.write(tl.generate_chrome_trace_format())
                        f.write('\n')

        #============================
        # Wrap tensorflow Session.run
        #============================

        self._should_trace = False
        self._run_metadata = RunMetaData()

        def _tfrunner(session, should_trace=False):
            """ Wrap the tensorflow Session.run method """
            trace_level = (tf.compat.v1.RunOptions.FULL_TRACE if should_trace
                                            else tf.compat.v1.RunOptions.NO_TRACE)
            options = tf.compat.v1.RunOptions(trace_level=trace_level)

            def _runner(*args, **kwargs):
                """ Pass options through """
                return session.run(*args, options=options, **kwargs)

            def _meta_runner(*args, **kwargs):
                """ Aggregate run metadata for each run """
                try:
                    run_metadata = tf.compat.v1.RunMetadata()
                    return session.run(*args, options=options,
                                              run_metadata=run_metadata,
                                            **kwargs)
                finally:
                    self._run_metadata.save(run_metadata)

            return _meta_runner if should_trace else _runner

        self._tfrun = _tfrunner(self._tf_session, self._should_trace)
        self._iterations = 0

    def _descriptor_feed(self):
        try:
            self._descriptor_feed_impl()
        except Exception as e:
            montblanc.log.exception("Descriptor Exception")
            raise

    def _descriptor_feed_impl(self):
        session = self._tf_session

        # Copy dimensions of the main cube
        cube = self.hypercube.copy()
        LSA = self._tf_feed_data.local

        # Get space of iteration
        iter_args = _iter_args(self._iter_dims, cube)
        descriptors_fed = 0

        # Iterate through the hypercube space
        for i, iter_cube in enumerate(cube.cube_iter(*iter_args)):
            descriptor = self._transcoder.encode(iter_cube.dimensions(copy=False))
            feed_dict = {LSA.descriptor.placeholders[0] : descriptor }
            montblanc.log.debug('Encoding {i} {d}'.format(i=i, d=descriptor))
            session.run(LSA.descriptor.put_op, feed_dict=feed_dict)
            descriptors_fed += 1

        montblanc.log.info("Done feeding {n} descriptors.".format(
            n=descriptors_fed))

        feed_dict = {LSA.descriptor.placeholders[0] : [-1] }
        session.run(LSA.descriptor.put_op, feed_dict=feed_dict)

    def _feed(self, cube, data_sources, data_sinks, global_iter_args):
        """ Feed stub """
        try:
            self._feed_impl(cube, data_sources, data_sinks, global_iter_args)
        except Exception as e:
            montblanc.log.exception("Feed Exception")
            raise

    def _feed_impl(self, cube, data_sources, data_sinks, global_iter_args):
        """ Implementation of staging_area feeding """
        session = self._tf_session
        FD = self._tf_feed_data
        LSA = FD.local

        # Get source strides out before the local sizes are modified during
        # the source loops below
        src_types = list(LSA.sources.keys())
        src_strides = [int(i) for i in cube.dim_extent_size(*src_types)]
        src_staging_areas = [[LSA.sources[t][s] for t in src_types]
            for s in range(self._nr_of_shards)]

        compute_feed_dict = { ph: cube.dim_global_size(n) for
            n, ph in list(FD.src_ph_vars.items()) }
        compute_feed_dict.update({ ph: getattr(cube, n) for
            n, ph in list(FD.property_ph_vars.items()) })

        chunks_fed = 0

        which_shard = itertools.cycle([self._shard(d,s)
            for s in range(self._shards_per_device)
            for d, dev in enumerate(self._devices)])

        while True:
            try:
                # Get the descriptor describing a portion of the RIME
                result = session.run(LSA.descriptor.get_op)
                descriptor = result['descriptor']
            except tf.errors.OutOfRangeError as e:
                montblanc.log.exception("Descriptor reading exception")

            # Quit if EOF
            if descriptor[0] == -1:
                break

            # Make it read-only so we can hash the contents
            descriptor.flags.writeable = False

            # Find indices of the emptiest staging_areas and, by implication
            # the shard with the least work assigned to it
            shard = next(which_shard) # round robin cycle on shards
            # submit a job - once loaded the shard will be incremented for compute to start
            feed_f = self._feed_executors[shard].submit(self._feed_actual,
                data_sources.copy(), cube.copy(),
                descriptor, shard,
                src_types, src_strides, src_staging_areas[shard],
                global_iter_args)

            compute_f = self._compute_executors[shard].submit(self._compute,
                compute_feed_dict, shard)

            consume_f = self._consumer_executor.submit(self._consume,
                data_sinks.copy(), cube.copy(), global_iter_args)

            yield (feed_f, compute_f, consume_f)
            chunks_fed += 1
        montblanc.log.info("Done feeding {n} chunks.".format(n=chunks_fed))

    def _feed_actual(self, *args):
        try:
            return self._feed_actual_impl(*args)
        except Exception as e:
            montblanc.log.exception("Feed Exception")
            raise

    def _feed_actual_impl(self, data_sources, cube,
            descriptor, shard,
            src_types, src_strides, src_staging_areas,
            global_iter_args):

        session = self._tf_session
        iq = self._tf_feed_data.local.feed_many[shard]

        # Decode the descriptor and update our cube dimensions
        dims = self._transcoder.decode(descriptor)
        cube.update_dimensions(dims)

        # Determine array shapes and data types for this
        # portion of the hypercube
        array_schemas = cube.arrays(reify=True)

        # Inject a data source and array schema for the
        # descriptor staging_area items.
        # These aren't full on arrays per se
        # but they need to work within the feeding framework
        array_schemas['descriptor'] = descriptor
        data_sources['descriptor'] = DataSource(
            lambda c: descriptor, np.int32, 'Internal')

        # Generate (name, placeholder, datasource, array schema)
        # for the arrays required by each staging_area
        gen = ((a, ph, data_sources[a], array_schemas[a])
            for ph, a in zip(iq.placeholders, iq.fed_arrays))

        # Get input data by calling the data source functors
        input_data = [(a, ph, _get_data(ds, SourceContext(a, cube,
                self.config(), global_iter_args,
                cube.array(a) if a in cube.arrays() else {},
                ad.shape, ad.dtype)))
            for (a, ph, ds, ad) in gen]

        # Create a feed dictionary from the input data
        feed_dict = { ph: data for (a, ph, data) in input_data }

        # Cache the inputs for this chunk of data,
        # so that sinks can access them
        input_cache = { a: data for (a, ph, data) in input_data }
        self._source_cache[descriptor.data] = input_cache

        montblanc.log.info("Enqueueing chunk {d} on shard {sh}".format(
            d=descriptor, sh=shard))

        self._tfrun(iq.put_op, feed_dict=feed_dict)

        # For each source type, feed that source staging_area
        for src_type, staging_area, stride in zip(src_types, src_staging_areas, src_strides):
            iter_args = [(src_type, stride)]

            # Iterate over chunks of the source
            for chunk_i, dim_desc in enumerate(cube.dim_iter(*iter_args)):
                cube.update_dimensions(dim_desc)
                s = dim_desc[0]['upper_extent'] - dim_desc[0]['lower_extent']


                montblanc.log.info("'{ci}: Enqueueing {d} '{s}' '{t}' sources "
                    "on shard {sh}".format(d=descriptor,
                        ci=chunk_i, s=s, t=src_type, sh=shard))

                # Determine array shapes and data types for this
                # portion of the hypercube
                array_schemas = cube.arrays(reify=True)

                # Generate (name, placeholder, datasource, array descriptor)
                # for the arrays required by each staging_area
                gen = [(a, ph, data_sources[a], array_schemas[a])
                    for ph, a in zip(staging_area.placeholders, staging_area.fed_arrays)]

                # Create a feed dictionary by calling the data source functors
                feed_dict = { ph: _get_data(ds, SourceContext(a, cube,
                        self.config(), global_iter_args + iter_args,
                        cube.array(a) if a in cube.arrays() else {},
                        ad.shape, ad.dtype))
                    for (a, ph, ds, ad) in gen }

                self._tfrun(staging_area.put_op, feed_dict=feed_dict)
        
        # input finally loaded now increment the number of inputs waiting on the shard
        self._inputs_waiting.increment(shard)

    def _compute(self, feed_dict, shard):
        """ Call the tensorflow compute """

        try:
            with self._inputs_waiting.getlock(shard):
                # only execute once data is indicated to be loaded for
                # this shard. Otherwise wait for the next iteration of the
                # main event loop to request another compute on this shard
                if self._inputs_waiting.get(shard, skiplock=True) > 0:
                    descriptor, enq = self._tfrun(self._tf_expr[shard], feed_dict=feed_dict)
                    self._inputs_waiting.decrement(shard, skiplock=True)
                else: #postpone work till data becomes available
                    self._compute_executors[shard].submit(self._compute,
                                                          feed_dict, shard)

        except Exception as e:
            montblanc.log.exception("Compute Exception")
            raise


    def _consume(self, data_sinks, cube, global_iter_args):
        """ Consume stub """
        try:
            return self._consume_impl(data_sinks, cube, global_iter_args)
        except Exception as e:
            montblanc.log.exception("Consumer Exception")
            six.reraise(Exception, Exception(e), sys.exc_info()[2])

    def _consume_impl(self, data_sinks, cube, global_iter_args):
        """ Consume """

        LSA = self._tf_feed_data.local
        output = self._tfrun(LSA.output.get_op)

        # Expect the descriptor in the first tuple position
        assert len(output) > 0
        assert LSA.output.fed_arrays[0] == 'descriptor'

        descriptor = output['descriptor']
        # Make it read-only so we can hash the contents
        descriptor.flags.writeable = False

        dims = self._transcoder.decode(descriptor)
        cube.update_dimensions(dims)

        # Obtain and remove input data from the source cache
        try:
            input_data = self._source_cache.pop(descriptor.data)
        except KeyError:
            raise ValueError("No input data cache available "
                "in source cache for descriptor {}!"
                    .format(descriptor))

        # For each array in our output, call the associated data sink
        gen = ((n, a) for n, a in list(output.items()) if not n == 'descriptor')

        for n, a in gen:
            sink_context = SinkContext(n, cube,
                self.config(), global_iter_args,
                cube.array(n) if n in cube.arrays() else {},
                a, input_data)

            _supply_data(data_sinks[n], sink_context)

    def solve(self, *args, **kwargs):
        #  Obtain source and sink providers, including internal providers
        source_providers = (self._source_providers +
            kwargs.get('source_providers', []))
        sink_providers = (self._sink_providers +
            kwargs.get('sink_providers', []))

        src_provs_str = 'Source Providers ' + str([sp.name() for sp
                                                in source_providers])
        snk_provs_str = 'Sink Providers ' + str([sp.name() for sp
                                                in sink_providers])

        montblanc.log.info(src_provs_str)
        montblanc.log.info(snk_provs_str)

        # Allow providers to initialise themselves based on
        # the given configuration
        ctx = InitialisationContext(self.config())

        for p in itertools.chain(source_providers, sink_providers):
            p.init(ctx)

        # Apply any dimension updates from the source provider
        # to the hypercube, taking previous reductions into account
        bytes_required = _apply_source_provider_dim_updates(
            self.hypercube, source_providers,
            self._previous_budget_dims)

        # If we use more memory than previously,
        # perform another budgeting operation
        # to make sure everything fits
        if bytes_required > self._previous_budget:
            self._previous_budget_dims, self._previous_budget = (
                _budget(self.hypercube, self.config()))

        # Determine the global iteration arguments
        # e.g. [('ntime', 100), ('nbl', 20)]
        global_iter_args = _iter_args(self._iter_dims, self.hypercube)

        # Indicate solution started in providers
        ctx = StartContext(self.hypercube, self.config(), global_iter_args)

        for p in itertools.chain(source_providers, sink_providers):
            p.start(ctx)

        #===================================
        # Assign data to Feed Once variables
        #===================================

        # Copy the hypercube
        cube = self.hypercube.copy()
        array_schemas = cube.arrays(reify=True)

        # Construct data sources from those supplied by the
        # source providers, if they're associated with
        # input sources
        LSA = self._tf_feed_data.local
        input_sources = LSA.input_sources
        data_sources = {n: DataSource(f, cube.array(n).dtype, prov.name())
            for prov in source_providers
            for n, f in list(prov.sources().items())
            if n in input_sources}

        # Get data sinks from supplied providers
        data_sinks = { n: DataSink(f, prov.name())
            for prov in sink_providers
            for n, f in list(prov.sinks().items())
            if not n == 'descriptor' }

        # Construct a feed dictionary from data sources
        feed_dict = {  fo.ph: _get_data(data_sources[k],
                SourceContext(k, cube,
                    self.config(), global_iter_args,
                    cube.array(k) if k in cube.arrays() else {},
                    array_schemas[k].shape,
                    array_schemas[k].dtype))
            for k, fo
            in list(LSA.feed_once.items()) }

        self._run_metadata.clear()

        # Run the assign operations for each feed_once variable
        assign_ops = [fo.assign_op.op for fo in list(LSA.feed_once.values())]
        self._tfrun(assign_ops, feed_dict=feed_dict)

        try:
            # Run the descriptor executor immediately
            params = self._descriptor_executor.submit(self._descriptor_feed)

            # Sets to track futures not yet completed
            feed_not_done = set()
            compute_not_done = set([params])
            consume_not_done = set()
            throttle_factor = self._nr_of_shards*QUEUE_SIZE

            # _feed_impl generates 3 futures
            # one for feeding data, one for computing with this data
            # and another for consuming it.
            # Iterate over these futures
            for feed, compute, consume in self._feed_impl(cube,
                data_sources, data_sinks, global_iter_args):

                feed_not_done.add(feed)
                compute_not_done.add(compute)
                consume_not_done.add(consume)

                # If there are many feed futures in flight,
                # perform throttling
                if len(feed_not_done) > throttle_factor*2:
                    # Wait for throttle_factor futures to complete
                    fit = cf.as_completed(feed_not_done)
                    feed_done = set(itertools.islice(fit, throttle_factor))
                    feed_not_done.difference_update(feed_done)

                    # Take an completed compute and consume
                    # futures immediately
                    compute_done, compute_not_done = cf.wait(
                        compute_not_done, timeout=0,
                        return_when=cf.FIRST_COMPLETED)
                    consume_done, consume_not_done = cf.wait(
                        consume_not_done, timeout=0,
                        return_when=cf.FIRST_COMPLETED)

                    # Get future results, mainly to fire exceptions
                    for i, f in enumerate(itertools.chain(feed_done,
                                        compute_done, consume_done)):
                        f.result()

                    not_done = sum(len(s) for s in (feed_not_done,
                        compute_not_done, consume_not_done))

                    montblanc.log.debug("Consumed {} futures. "
                        "{} remaining".format(i, not_done))

            # Request future results, mainly for exceptions
            for f in cf.as_completed(itertools.chain(feed_not_done,
                    compute_not_done, consume_not_done)):

                f.result()

        except (KeyboardInterrupt, SystemExit) as e:
            montblanc.log.exception('Solving interrupted')
            raise
        except Exception:
            montblanc.log.exception('Solving exception')
            raise
        else:
            if self._should_trace:
                self._run_metadata.write(self._iterations)

            self._iterations += 1
        finally:
            # Indicate solution stopped in providers
            ctx = StopContext(self.hypercube, self.config(), global_iter_args)
            for p in itertools.chain(source_providers, sink_providers):
                p.stop(ctx)

            montblanc.log.info('Solution Completed')


    def close(self):
        # Shutdown thread executors
        self._descriptor_executor.shutdown()
        [fe.shutdown() for fe in self._feed_executors]
        [ce.shutdown() for ce in self._compute_executors]
        self._consumer_executor.shutdown()

        # Shutdown thte tensorflow session
        self._tf_session.close()

        # Shutdown data sources
        for source in self._source_providers:
            source.close()

        # Shutdown data sinks
        for sink in self._sink_providers:
            sink.close()

    def __enter__(self):
        return self

    def __exit__(self, etype, evalue, etrace):
        self.close()


def _create_defaults_source_provider(cube, data_source):
    """
    Create a DefaultsSourceProvider object. This provides default
    data sources for each array defined on the hypercube. The data sources
    may either by obtained from the arrays 'default' data source
    or the 'test' data source.
    """
    from montblanc.impl.rime.tensorflow.sources import (
        find_sources, DEFAULT_ARGSPEC)
    from montblanc.impl.rime.tensorflow.sources import constant_cache

    # Obtain default data sources for each array,
    # Just take from defaults if test data isn't specified
    staging_area_data_source = ('default' if not data_source == 'test'
                                                      else data_source)

    cache = True

    default_prov = DefaultsSourceProvider(cache=cache)

    # Create data sources on the source provider from
    # the cube array data sources
    for n, a in list(cube.arrays().items()):
        # Unnecessary for temporary arrays
        if 'temporary' in a.tags:
            continue

        # Obtain the data source
        data_source = a.get(staging_area_data_source)

        # Array marked as constant, decorate the data source
        # with a constant caching decorator
        if cache is True and 'constant' in a.tags:
            data_source = constant_cache(data_source)

        method = types.MethodType(data_source, default_prov)
        setattr(default_prov, n, method)

    def _sources(self):
        """
        Override the sources method to also handle lambdas that look like
        lambda s, c: ..., as defined in the config module
        """

        try:
            return self._sources
        except AttributeError:
            self._sources = find_sources(self, [DEFAULT_ARGSPEC] + [['s', 'c']])

        return self._sources

    # Monkey patch the sources method
    default_prov.sources = types.MethodType(_sources, default_prov)

    return default_prov

def _construct_tensorflow_feed_data(dfs, cube, iter_dims,
    nr_of_input_staging_areas):

    FD = AttrDict()
    FD['_sequence_type'] = list
    # Reference local staging_areas
    FD.local = local = AttrDict()
    local['_sequence_type'] = list

    # Create placholder variables for source counts
    FD.src_ph_vars = AttrDict({
        n: tf.compat.v1.placeholder(dtype=tf.int32, shape=(), name=n)
        for n in ['nsrc'] + mbu.source_nr_vars()})

    # Create placeholder variables for properties
    FD.property_ph_vars = AttrDict({
        n: tf.compat.v1.placeholder(dtype=p.dtype, shape=(), name=n)
        for n, p in list(cube.properties().items()) })

    #========================================================
    # Determine which arrays need feeding once/multiple times
    #========================================================

    # Take all arrays flagged as input
    input_arrays = [a for a in list(cube.arrays().values())
                    if 'input' in a.tags]

    src_data_sources, feed_many, feed_once = _partition(iter_dims,
                                                        input_arrays)

    #=====================================
    # Descriptor staging area
    #=====================================

    local.descriptor = create_staging_area_wrapper('descriptors',
        ['descriptor'], dfs)

    #===========================================
    # Staging area for multiply fed data sources
    #===========================================

    # Create the staging_area for holding the feed many input
    local.feed_many = [create_staging_area_wrapper('feed_many_%d' % i,
                ['descriptor'] + [a.name for a in feed_many], dfs)
            for i in range(nr_of_input_staging_areas)]

    #=================================================
    # Staging areas for each radio source data sources
    #=================================================

    # Create the source array staging areas
    local.sources = { src_nr_var: [
            create_staging_area_wrapper('%s_%d' % (src_type, i),
            [a.name for a in src_data_sources[src_nr_var]], dfs)
            for i in range(nr_of_input_staging_areas)]

        for src_type, src_nr_var in list(source_var_types().items())
    }

    #======================================
    # The single output staging_area
    #======================================

    local.output = create_staging_area_wrapper('output',
        ['descriptor', 'model_vis', 'chi_squared'], dfs)

    #=================================================
    # Create tensorflow variables which are
    # fed only once via an assign operation
    #=================================================

    def _make_feed_once_tuple(array):
        name = array.name
        dtype = dfs[name].dtype
        shape = tuple([None if isinstance(x, str) else x 
                       for x in array.shape])
        ph = tf.compat.v1.placeholder(dtype=dtype,
            name=name + "_placeholder", shape=shape)

        var = tf.compat.v1.Variable(ph, 
                                    name=name)

        op = tf.compat.v1.assign(var, ph)

        return FeedOnce(ph, var, op)

    # Create placeholders, variables and assign operators
    # for data sources that we will only feed once
    local.feed_once = { a.name : _make_feed_once_tuple(a)
        for a in feed_once }

    #=======================================================
    # Construct the list of data sources that need feeding
    #=======================================================

    # Data sources from input staging_areas
    src_sa = [q for sq in list(local.sources.values()) for q in sq]
    all_staging_areas = local.feed_many + src_sa
    input_sources = { a for q in all_staging_areas
                        for a in q.fed_arrays}
    # Data sources from feed once variables
    input_sources.update(list(local.feed_once.keys()))

    local.input_sources = input_sources

    return FD

def _construct_tensorflow_expression(slvr_cfg, feed_data, device, shard):
    """ Constructs a tensorflow expression for computing the RIME """
    zero = tf.constant(0)
    src_count = zero
    src_ph_vars = feed_data.src_ph_vars

    LSA = feed_data.local

    polarisation_type = slvr_cfg['polarisation_type']

    # Pull RIME inputs out of the feed staging_area
    # of the relevant shard, adding the feed once
    # inputs to the dictionary
    D = LSA.feed_many[shard].get_to_attrdict()
    D.update({k: fo.var for k, fo in list(LSA.feed_once.items())})

    with tf.device(device):
        # Infer chunk dimensions
        model_vis_shape = tf.shape(input=D.model_vis)
        ntime, nbl, nchan, npol = [model_vis_shape[i] for i in range(4)]

        # Infer float and complex type
        FT, CT = D.uvw.dtype, D.model_vis.dtype

        # Compute sine and cosine of parallactic angles
        pa_sin, pa_cos = rime.parallactic_angle_sin_cos(
				D.parallactic_angles[:, :] + 
				D.feed_angles[None, :])
        # Compute feed rotation
        feed_rotation = rime.feed_rotation(pa_sin, pa_cos, CT=CT,
                                           feed_type=polarisation_type)

    def antenna_jones(radec, stokes):
        """
        Compute the jones terms for each antenna.

        lm, stokes and alpha are the source variables.
        """
        lm = rime.radec_to_lm(radec, D.phase_centre)

        # Compute the complex phase
        cplx_phase = rime.phase(lm, D.uvw, D.frequency, CT=CT)

        # Check for nans/infs in the complex phase
        phase_msg = ("Check that '1 - l**2  - m**2 >= 0' holds "
                     "for all your lm coordinates. This is required "
                     "for 'n = sqrt(1 - l**2 - m**2) - 1' "
                     "to be finite.")

        phase_real = tf.debugging.check_numerics(tf.math.real(cplx_phase), phase_msg)
        phase_imag = tf.debugging.check_numerics(tf.math.imag(cplx_phase), phase_msg)

        # Compute the square root of the brightness matrix
        # (as well as the sign)
        bsqrt, sgn_brightness = rime.b_sqrt(stokes, CT=CT,
            polarisation_type=polarisation_type)

        # Check for nans/infs in the bsqrt
        bsqrt_msg = ("Check that your stokes parameters "
                     "satisfy I**2 >= Q**2 + U**2 + V**2. "
                     "Montblanc performs a cholesky decomposition "
                     "of the brightness matrix and the above must "
                     "hold for this to produce valid values.")

        bsqrt_real = tf.debugging.check_numerics(tf.math.real(bsqrt), bsqrt_msg)
        bsqrt_imag = tf.debugging.check_numerics(tf.math.imag(bsqrt), bsqrt_msg)

        # Compute the direction dependent effects from the beam
        #radec_prime = radec * tf.cast(tf.stack([-1.0, 1.0]), radec.dtype)
        #phase_centre_prime = D.phase_centre * tf.cast(tf.stack([-1.0, 1.0]), D.phase_centre.dtype)
        #def normang(val):
        #    """ Normalizes angle between [-pi, pi) """
        #    return ( val + np.pi) % ( 2 * np.pi ) - np.pi
        #cube_pos = normang(normang(radec_prime) - normang(phase_centre_prime))

        ejones = rime.e_beam(lm, D.frequency,
                             D.pointing_errors, D.antenna_scaling,
                             pa_sin, pa_cos,
                             D.beam_extents, D.beam_freq_map, D.ebeam)

        deps = [phase_real, phase_imag, bsqrt_real, bsqrt_imag]
        deps = [] # Do nothing for now

        # Combine the brightness square root, complex phase,
        # feed rotation and beam dde's
        with tf.control_dependencies(deps):
            antenna_jones = rime.create_antenna_jones(bsqrt, cplx_phase,
                                                      feed_rotation, ejones,
                                                      FT=FT)
            return antenna_jones, sgn_brightness

    # While loop condition for each point source type
    def point_cond(coherencies, npsrc, src_count):
        return tf.less(npsrc, src_ph_vars.npsrc)

    def gaussian_cond(coherencies, ngsrc, src_count):
        return tf.less(ngsrc, src_ph_vars.ngsrc)

    def sersic_cond(coherencies, nssrc, src_count):
        return tf.less(nssrc, src_ph_vars.nssrc)

    # While loop bodies
    def point_body(coherencies, npsrc, src_count):
        """ Accumulate visiblities for point source batch """
        S = LSA.sources['npsrc'][shard].get_to_attrdict()

        # Maintain source counts
        nsrc = tf.shape(input=S.point_lm)[0]
        src_count += nsrc
        npsrc +=  nsrc

        ant_jones, sgn_brightness = antenna_jones(S.point_lm,
            S.point_stokes)
        shape = tf.ones(shape=[nsrc,ntime,nbl,nchan], dtype=FT)
        coherencies = rime.sum_coherencies(D.antenna1, D.antenna2,
            shape, ant_jones, sgn_brightness, coherencies)

        return coherencies, npsrc, src_count

    def gaussian_body(coherencies, ngsrc, src_count):
        """ Accumulate coherencies for gaussian source batch """
        S = LSA.sources['ngsrc'][shard].get_to_attrdict()

        # Maintain source counts
        nsrc = tf.shape(input=S.gaussian_lm)[0]
        src_count += nsrc
        ngsrc += nsrc

        ant_jones, sgn_brightness = antenna_jones(S.gaussian_lm,
            S.gaussian_stokes)
        gauss_shape = rime.gauss_shape(D.uvw, D.antenna1, D.antenna2,
            D.frequency, S.gaussian_shape)
        coherencies = rime.sum_coherencies(D.antenna1, D.antenna2,
            gauss_shape, ant_jones, sgn_brightness, coherencies)

        return coherencies, ngsrc, src_count

    def sersic_body(coherencies, nssrc, src_count):
        """ Accumulate coherencies for sersic source batch """
        S = LSA.sources['nssrc'][shard].get_to_attrdict()

        # Maintain source counts
        nsrc = tf.shape(input=S.sersic_lm)[0]
        src_count += nsrc
        nssrc += nsrc

        ant_jones, sgn_brightness = antenna_jones(S.sersic_lm,
            S.sersic_stokes)
        sersic_shape = rime.sersic_shape(D.uvw, D.antenna1, D.antenna2,
            D.frequency, S.sersic_shape)
        coherencies = rime.sum_coherencies(D.antenna1, D.antenna2,
            sersic_shape, ant_jones, sgn_brightness, coherencies)

        return coherencies, nssrc, src_count

    with tf.device(device):
        base_coherencies = tf.zeros(shape=[ntime,nbl,nchan,npol], dtype=CT)

        # Evaluate point sources
        summed_coherencies, npsrc, src_count = tf.while_loop(
            cond=point_cond, body=point_body,
            loop_vars=[base_coherencies, zero, src_count])

        # Evaluate gaussians
        summed_coherencies, ngsrc, src_count = tf.while_loop(
            cond=gaussian_cond, body=gaussian_body,
            loop_vars=[summed_coherencies, zero, src_count])

        # Evaluate sersics
        summed_coherencies, nssrc, src_count = tf.while_loop(
            cond=sersic_cond, body=sersic_body,
            loop_vars=[summed_coherencies, zero, src_count])

        # Post process visibilities to produce model visibilites and chi squared
        model_vis, chi_squared = rime.post_process_visibilities(
            D.antenna1, D.antenna2, D.direction_independent_effects, D.flag,
            D.weight, D.model_vis, summed_coherencies, D.observed_vis)

    # Create enstaging_area operation
    put_op = LSA.output.put_from_list([D.descriptor, model_vis, chi_squared])

    # Return descriptor and enstaging_area operation
    return D.descriptor, put_op

def _get_data(data_source, context):
    """ Get data from the data source, checking the return values """
    try:
        # Get data from the data source
        data = data_source.source(context)

        # Complain about None values
        if data is None:
            raise ValueError("'None' returned from "
                "data source '{n}'".format(n=context.name))
        # We want numpy arrays
        elif not isinstance(data, np.ndarray):
            raise TypeError("Data source '{n}' did not "
                "return a numpy array, returned a '{t}'".format(
                    t=type(data)))
        # And they should be the right shape and type
        elif data.shape != context.shape or data.dtype != context.dtype:
            raise ValueError("Expected data of shape '{esh}' and "
                "dtype '{edt}' for data source '{n}', but "
                "shape '{rsh}' and '{rdt}' was found instead".format(
                    n=context.name, esh=context.shape, edt=context.dtype,
                    rsh=data.shape, rdt=data.dtype))

        return data

    except Exception as e:
        ex = ValueError("An exception occurred while "
            "obtaining data from data source '{ds}'\n\n"
            "{e}\n\n"
            "{help}".format(ds=context.name,
                e=str(e), help=context.help()))

        six.reraise(ValueError, ex, sys.exc_info()[2])

def _supply_data(data_sink, context):
    """ Supply data to the data sink """
    try:
        data_sink.sink(context)
    except Exception as e:
        ex = ValueError("An exception occurred while "
            "supplying data to data sink '{ds}'\n\n"
            "{e}\n\n"
            "{help}".format(ds=context.name,
                e=str(e), help=context.help()))

        six.reraise(ValueError, ex, sys.exc_info()[2])

def _iter_args(iter_dims, cube):
    iter_strides = cube.dim_extent_size(*iter_dims)
    return list(zip(iter_dims, iter_strides))

def _uniq_log2_range(start, size, div):
    start = np.log2(start)
    size = np.log2(size)
    int_values = np.int32(np.logspace(start, size, div, base=2)[:-1])

    return np.flipud(np.unique(int_values))

def _budget(cube, slvr_cfg):
    # Figure out a viable dimension configuration
    # given the total problem size
    mem_budget = slvr_cfg.get('mem_budget', 2*ONE_GB)
    bytes_required = cube.bytes_required()

    src_dims = mbu.source_nr_vars() + ['nsrc']
    dim_names = ['na', 'nbl', 'ntime'] + src_dims
    global_sizes = cube.dim_global_size(*dim_names)
    na, nbl, ntime = global_sizes[:3]

    # Keep track of original dimension sizes and any reductions that are applied
    original_sizes = { r: s for r, s in zip(dim_names, global_sizes) }
    applied_reductions = {}

    def _reduction():
        # Reduce over time first
        trange = _uniq_log2_range(1, ntime, 5)
        for t in trange[0:1]:
            yield [('ntime', t)]

        # Attempt reduction over source
        sbs = slvr_cfg['source_batch_size']
        srange = _uniq_log2_range(10, sbs, 5) if sbs > 10 else 10
        src_dim_gs = global_sizes[3:]

        for bs in srange:
            yield [(d, bs if bs < gs else gs) for d, gs
                in zip(src_dims, src_dim_gs)]

        # Try the rest of the timesteps
        for t in trange[1:]:
            yield [('ntime', t)]

        # Reduce by baseline
        for bl in _uniq_log2_range(na, nbl, 5):
            yield [('nbl', bl)]

    for reduction in _reduction():
        if bytes_required > mem_budget:
            for dim, size in reduction:
                applied_reductions[dim] = size
                cube.update_dimension(dim, lower_extent=0, upper_extent=size)
        else:
            break

        bytes_required = cube.bytes_required()

    # Log some information about the memory_budget
    # and dimension reduction
    montblanc.log.info(("Selected a solver memory budget of {rb} "
        "given a hard limit of {mb}.").format(
        rb=mbu.fmt_bytes(bytes_required),
        mb=mbu.fmt_bytes(mem_budget)))

    if len(applied_reductions) > 0:
        montblanc.log.info("The following dimension reductions "
            "were applied:")

        for k, v in list(applied_reductions.items()):
            montblanc.log.info('{p}{d}: {id} => {rd}'.format
                (p=' '*4, d=k, id=original_sizes[k], rd=v))
    else:
        montblanc.log.info("No dimension reductions were applied.")

    return applied_reductions, bytes_required

DimensionUpdate = attr.make_class("DimensionUpdate",
    ['size', 'prov'], slots=True, frozen=True)

def _apply_source_provider_dim_updates(cube, source_providers, budget_dims):
    """
    Given a list of source_providers, apply the list of
    suggested dimension updates given in provider.updated_dimensions()
    to the supplied hypercube.

    Dimension global_sizes are always updated with the supplied sizes and
    lower_extent is always set to 0. upper_extent is set to any reductions
    (current upper_extents) existing in budget_dims, otherwise it is set
    to global_size.

    """
    # Create a mapping between a dimension and a
    # list of (global_size, provider_name) tuples
    update_map = collections.defaultdict(list)

    for prov in source_providers:
        for dim_tuple in prov.updated_dimensions():
            name, size = dim_tuple

            # Don't accept any updates on the nsrc dimension
            # This is managed internally
            if name == 'nsrc':
                continue

            dim_update = DimensionUpdate(size, prov.name())
            update_map[name].append(dim_update)

    # No dimensions were updated, quit early
    if len(update_map) == 0:
        return cube.bytes_required()

    # Ensure that the global sizes we receive
    # for each dimension are unique. Tell the user
    # when conflicts occur
    update_list = []

    for name, updates in list(update_map.items()):
        if not all(updates[0].size == du.size for du in updates[1:]):
            raise ValueError("Received conflicting "
                "global size updates '{u}'"
                " for dimension '{n}'.".format(n=name, u=updates))

        update_list.append((name, updates[0].size))

    montblanc.log.info("Updating dimensions {} from "
                        "source providers.".format(str(update_list)))

    # Now update our dimensions
    for name, global_size in update_list:
        # Defer to existing any existing budgeted extent sizes
        # Otherwise take the global_size
        extent_size = budget_dims.get(name, global_size)

        # Take the global_size if extent_size was previously zero!
        extent_size = global_size if extent_size == 0 else extent_size

        # Clamp extent size to global size
        if extent_size > global_size:
            extent_size = global_size

        # Update the dimension
        cube.update_dimension(name,
            global_size=global_size,
            lower_extent=0,
            upper_extent=extent_size)

    # Handle global number of sources differently
    # It's equal to the number of
    # point's, gaussian's, sersic's combined
    nsrc = sum(cube.dim_global_size(*mbu.source_nr_vars()))

    # Extent size will be equal to whatever source type
    # we're currently iterating over. So just take
    # the maximum extent size given the sources
    es = max(cube.dim_extent_size(*mbu.source_nr_vars()))

    cube.update_dimension('nsrc',
        global_size=nsrc,
        lower_extent=0,
        upper_extent=es)

    # Return our cube size
    return cube.bytes_required()

def _setup_hypercube(cube, slvr_cfg):
    """ Sets up the hypercube given a solver configuration """
    mbu.register_default_dimensions(cube, slvr_cfg)

    # Configure the dimensions of the beam cube
    cube.register_dimension('beam_lw', 2,
                            description='E Beam cube l width')

    cube.register_dimension('beam_mh', 2,
                            description='E Beam cube m height')

    cube.register_dimension('beam_nud', 2,
                            description='E Beam cube nu depth')

    # =========================================
    # Register hypercube Arrays and Properties
    # =========================================

    from montblanc.impl.rime.tensorflow.config import (A, P)

    def _massage_dtypes(A, T):
        def _massage_dtype_in_dict(D):
            new_dict = D.copy()
            new_dict['dtype'] = mbu.dtype_from_str(D['dtype'], T)
            return new_dict

        return [_massage_dtype_in_dict(D) for D in A]

    dtype = slvr_cfg['dtype']
    is_f32 = dtype == 'float'

    T = {
        'ft' : np.float32 if is_f32 else np.float64,
        'ct' : np.complex64 if is_f32 else np.complex128,
        'int' : int,
    }

    cube.register_properties(_massage_dtypes(P, T))
    cube.register_arrays(_massage_dtypes(A, T))

def _partition(iter_dims, data_sources):
    """
    Partition data sources into

    1. Dictionary of data sources associated with radio sources.
    2. List of data sources to feed multiple times.
    3. List of data sources to feed once.
    """

    src_nr_vars = set(source_var_types().values())
    iter_dims = set(iter_dims)

    src_data_sources = collections.defaultdict(list)
    feed_many = []
    feed_once = []

    for ds in data_sources:
        # Is this data source associated with
        # a radio source (point, gaussian, etc.?)
        src_int = src_nr_vars.intersection(ds.shape)

        if len(src_int) > 1:
            raise ValueError("Data source '{}' contains multiple "
                            "source types '{}'".format(ds.name, src_int))
        elif len(src_int) == 1:
            # Yep, record appropriately and iterate
            src_data_sources[src_int.pop()].append(ds)
            continue

        # Are we feeding this data source multiple times
        # (Does it possess dimensions on which we iterate?)
        if len(iter_dims.intersection(ds.shape)) > 0:
            feed_many.append(ds)
            continue

        # Assume this is a data source that we only feed once
        feed_once.append(ds)

    return src_data_sources, feed_many, feed_once



import argparse
import fcntl
import json
import logging
import os
import select
import socket
import struct

class SyncError(Exception):
    pass

PING = 'PING'
PONG = 'PONG'

logging.basicConfig(level=logging.DEBUG, format='%(levelname)s - %(message)s')

parser = argparse.ArgumentParser()
parser.add_argument('-i', '--interface', default='eth0')
parser.add_argument('-p', '--port', default=8888)
parser.add_argument('--no-start', dest='start', action='store_false')
parser.add_argument('--start', dest='start', action='store_true')
parser.set_defaults(start=False)
args = parser.parse_args()

def get_ip_address(ifname):
    """ Hack to get IP address from the interface """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    return socket.inet_ntoa(fcntl.ioctl(
        s.fileno(),
        0x8915,  # SIOCGIFADDR
        struct.pack('256s', ifname[:15])
    )[20:24])

# Track client connections
connections = {}
lost = set()

# Determine host address to bind a server socket to
host_address = (get_ip_address(args.interface), args.port)

try:
    logging.info('Server listening on {a}'.format(a=host_address))

    host_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    host_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    host_socket.bind(host_address)
    host_socket.listen(5)

    while True:
        client_socket, client_address = host_socket.accept()
        logging.info('Connection from {a}'.format(a=client_address))
        connections[client_socket] = (client_socket, client_address)

except KeyboardInterrupt:
    logging.info('Ctrl^C received')

    logging.info('Pinging {n} connection(s)'.format(n=len(connections)))

    for k, (cs, ca) in list(connections.items()):
        try:
            cs.send(PING)
        except socket.error as e:
            logging.warn('Lost connection to {a}'.format(a=ca))
            lost.add((cs,ca))

    for k, (cs, ca) in list(connections.items()):
        try:
            if cs.recv(len(PONG)) != PONG:
                raise SyncError()
        except (socket.error, SyncError) as e:
            logging.warn('Lost connection to {a}'.format(a=ca))
            lost.add((cs, ca))

    logging.info('Lost {n} connection(s)'.format(n=len(lost)))

    connections = { k : c for k, c in list(connections.items())
        if c not in lost }

    logging.info('Creating cluster specification for {n} workers'.format(
        n=len(connections)))
    # Create the lists of workers and master urls
    master_list = ['{ip}:{port}'.format(ip=host_address[0], port=host_address[1])]
    worker_list = ['{ip}:{port}'.format(ip=ip, port=port) for (ip, port) in
        (s.getpeername() for s, _ in list(connections.values()))]

    logging.info('Master node(s) {n}'.format(n=master_list))
    logging.info('Worker node(s) {n}'.format(n=worker_list))

    cluster = { 'worker' : worker_list, 'master' : master_list }

    # Transmit cluster specification to connected clients
    for i, (cs, ca) in enumerate(connections.values()):
        data = { 'cluster' : cluster, 'job' : 'worker', 'task' : i }

        logging.info('Sending specification to {ca}'.format(ca=ca))
        cs.send(json.dumps(data))

finally:
    # Close client sockets
    for cs, address in list(connections.values()):
        logging.info('Closing connection to {c}'.format(c=address))
        cs.shutdown(socket.SHUT_RDWR)
        cs.close()

    for cs, address in lost:
        logging.info('Closing connection to {c}'.format(c=address))
        cs.close()

    # Close server socket
    host_socket.shutdown(socket.SHUT_RDWR)
    host_socket.close()
    logging.info('Closing host socket {h}'.format(h=host_address))

logging.info("Cluster specification\n{c}".format(c=cluster))

if args.start is True:
    import tensorflow as tf
    import numpy as np
    import time

    server = tf.distribute.Server(cluster, job_name='master', task_index=0)
    logging.info("Server Target is '{st}'".format(st=server.target))

    values = []

    g = tf.Graph()

    with g.as_default():
        with tf.device('/job:master/task:0'):
            with tf.compat.v1.container('shared'):
                queue_in = tf.queue.FIFOQueue(10, [tf.int32],
                    name='queue_in',
                    shared_name='master_queue_in')

                queue_out = tf.queue.FIFOQueue(10, [tf.string],
                    name='queue_out',
                    shared_name='master_queue_out')

                tmp = tf.Variable([100, 1000], dtype=tf.int32, name='master_tmp')

                q1 = queue_in.enqueue(1, name='q1')
                q2 = queue_in.enqueue(2, name='q2')
                q3 = queue_in.enqueue(3, name='q3')
                q4 = queue_in.enqueue(4, name='q4')

                do_enq = tf.group(q4, q3, q2, q1)

        for t in range(len(cluster['worker'])):
            with tf.device('/job:worker/task:{t}'.format(t=t)):
                A = tf.Variable(tf.ones(shape=(10,), dtype=tf.float32), name='a')
                B = tf.Variable(tf.ones(shape=(10,), dtype=tf.float32), name='b')
                C = A + B*2
                values.append(C)

        init_op = tf.compat.v1.initialize_variables([A, B, tmp])

        result = tf.pack(values)

        do_deq = queue_out.dequeue()

    with tf.compat.v1.Session(server.target, graph=g) as S:
        S.run(init_op)
        S.run(do_enq)
        print(('Worker result', S.run(result)))
        print(('Dequeue result', S.run(do_deq)))

        time.sleep(2)



import itertools

import numpy as np

# These are hypercube dimension attributes
DEFAULT_SCHEMA = ['lower_extent', 'upper_extent', 'global_size']

class CubeDimensionTranscoder(object):
    """
    Small class for encoding and decoding hypercube dimensions
    into/from numpy arrays

    >>> import hypercube as hc
    >>> cube = hc.HyperCube()
    >>> cube.register_dimension('ntime', 100)
    >>> cube.register_dimension('na', 7)
    >>> cube.register_dimension('nchan', 128)

    >>> dimdesc = CubeDimensionTranscoder(['ntime', 'nchan'])
    >>> desc = dimdesc.encode(cube.dimensions())
    >>> print desc
    >>> time, chan = dimdesc.decode(desc)
    >>> print time, chan
    """
    def __init__(self, dimensions, schema=None):
        if schema is None:
            schema = DEFAULT_SCHEMA
        elif not all([s in DEFAULT_SCHEMA for s in schema]):
            raise ValueError("Schema '{s}' contains invalid attributes. "
                "Valid attributes are '{v}'".format(s=schema, v=DEFAULT_SCHEMA))

        self._dimensions = dimensions
        self._schema = tuple(schema)

    @property
    def dimensions(self):
        return self._dimensions

    @property
    def schema(self):
        return self._schema

    def encode(self, cube_dimensions):
        """
        Produces a numpy array of integers which encode
        the supplied cube dimensions.
        """
        return np.asarray([getattr(cube_dimensions[d], s)
            for d in self._dimensions
            for s in self._schema],
                dtype=np.int32)

    def decode(self, descriptor):
        """ Produce a list of dictionaries for each dimension in this transcoder """
        i = iter(descriptor)
        n = len(self._schema)

        # Add the name key to our schema
        schema = self._schema + ('name',)
        # For each dimensions, generator takes n items off iterator
        # wrapping the descriptor, making a tuple with the dimension
        # name appended
        tuple_gen = (tuple(itertools.islice(i, n)) + (d, )
            for d in self._dimensions)

        # Generate dictionary by mapping schema keys to generated tuples
        return [{ k: v for k, v in zip(schema, t) } for t in tuple_gen]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import logging
import unittest
import numpy as np
import tempfile
import time

import montblanc
import montblanc.src_types as mbs
import montblanc.util as mbu

class TestSourceUtils(unittest.TestCase):
    """
    TestSourceUtils class defining unit tests for
    montblanc's montblanc.src_type module
    """

    def setUp(self):
        """ Set up each test case """
        np.random.seed(int(time.time()) & 0xFFFFFFFF)
        montblanc.setup_test_logging()

    def tearDown(self):
        """ Tear down each test case """
        pass

    def test_source_range(self):
        """
        Test that, given a source range, the returned dictionary
        contains the source types within that range
        """
        src_dict = mbs.sources_to_nr_vars({'point':10, 'gaussian':20, 'sersic':40})

        D = mbs.source_range(0, 45, src_dict)
        self.assertTrue(D['npsrc'] == 10 and D['ngsrc'] == 20 and D['nssrc'] == 15)

        D = mbs.source_range(3, 45, src_dict)
        self.assertTrue(D['npsrc'] == 7 and D['ngsrc'] == 20 and D['nssrc'] == 15)

        D = mbs.source_range(13, 45, src_dict)
        self.assertTrue(D['npsrc'] == 0 and D['ngsrc'] == 17 and D['nssrc'] == 15)

        D = mbs.source_range(0, 3, src_dict)
        self.assertTrue(D['npsrc'] == 3 and D['ngsrc'] == 0 and D['nssrc'] == 0)

        D = mbs.source_range(15, 18, src_dict)
        self.assertTrue(D['ngsrc'] == 3)

        D = mbs.source_range(15, 70, src_dict)
        self.assertTrue(D['npsrc'] == 0 and D['ngsrc'] == 15 and D['nssrc'] == 40)

    def test_source_range_tuple(self):
        """
        Test that, given a source range, the returned dictionary
        contains the source types within that range
        """
        src_dict = mbs.sources_to_nr_vars({'point':10, 'gaussian':20, 'sersic':40})

        D = mbs.source_range_tuple(0, 45, src_dict)
        self.assertTrue(D['npsrc'] == (0, 10) and
            D['ngsrc'] == (0, 20) and
            D['nssrc'] == (0, 15))

        D = mbs.source_range_tuple(3, 45, src_dict)
        self.assertTrue(D['npsrc'] == (3, 10) and
            D['ngsrc'] == (0, 20) and
            D['nssrc'] == (0, 15))

        D = mbs.source_range_tuple(13, 45, src_dict)
        self.assertTrue(D['npsrc'] == (0, 0) and
            D['ngsrc'] == (3, 20) and
            D['nssrc'] == (0, 15))

        D = mbs.source_range_tuple(0, 3, src_dict)
        self.assertTrue(D['npsrc'] == (0, 3) and
            D['ngsrc'] == (0, 0) and
            D['nssrc'] == (0, 0))

        D = mbs.source_range_tuple(15, 18, src_dict)
        self.assertTrue(D['npsrc'] == (0, 0) and
            D['ngsrc'] == (5, 8) and
            D['nssrc'] == (0, 0))

        D = mbs.source_range_tuple(15, 70, src_dict)
        self.assertTrue(D['npsrc'] == (0, 0) and
            D['ngsrc'] == (5, 20) and
            D['nssrc'] == (0, 40))

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSourceUtils)
    unittest.TextTestRunner(verbosity=2).run(suite)



import itertools
import unittest

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tensorflow.python.client import device_lib

class TestPostProcessVisibilities(unittest.TestCase):
    """ Tests the PostProcessVisibilities operator """

    def setUp(self):
        # Load the rime operation library
        from montblanc.impl.rime.tensorflow import load_tf_lib
        self.rime = load_tf_lib()

        # Load the custom operation library
        #self.rime = tf.load_op_library('rime.so')
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                                if d.device_type == 'GPU']

    def test_post_process_visibilities(self):
        """ Test the PostProcessVisibilities operator """

        # List of type constraints for testing this operator
        type_permutations = [
            [np.float32, np.complex64],
            [np.float64, np.complex128]]

        # Run test with the type combinations above
        for FT, CT in type_permutations:
            self._impl_test_post_process_visibilities(FT, CT)

    def _impl_test_post_process_visibilities(self, FT, CT):
        """ Implementation of the PostProcessVisibilities operator test """

        ntime, nbl, na, nchan = 100, 21, 7, 16

        rf = lambda *a, **kw: np.random.random(*a, **kw).astype(FT)
        rc = lambda *a, **kw: rf(*a, **kw) + 1j*rf(*a, **kw).astype(CT)

        # Create input variables
        antenna1 = np.random.randint(low=0, high=na,
            size=[ntime, nbl]).astype(np.int32)
        antenna2 = np.random.randint(low=0, high=na,
            size=[ntime, nbl]).astype(np.int32)
        direction_independent_effects = rc(size=[ntime, na, nchan, 4])
        flag = np.random.randint(low=0, high=2,
            size=[ntime, nbl, nchan, 4]).astype(np.uint8)
        weight = rf(size=[ntime, nbl, nchan, 4])
        base_vis = rc(size=[ntime, nbl, nchan, 4])
        model_vis = rc(size=[ntime, nbl, nchan, 4])
        observed_vis = rc(size=[ntime, nbl, nchan, 4])

        # Argument list
        np_args = [antenna1, antenna2, direction_independent_effects, flag, weight,
            base_vis, model_vis, observed_vis]
        # Argument string name list
        arg_names = ['antenna1', 'antenna2', 'direction_independent_effects',
            'flag', 'weight', 'base_vis', 'model_vis', 'observed_vis']
        # Constructor tensorflow variables
        tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, arg_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return self.rime.post_process_visibilities(*tf_args)

        # Pin operation to CPU
        cpu_op = _pin_op('/cpu:0', *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.compat.v1.global_variables_initializer()

        with tf.compat.v1.Session() as S:
            S.run(init_op)

            # Get the CPU visiblities and chi squared
            cpu_vis, cpu_X2 = S.run(cpu_op)

            # Compare against the gpu visibilities and chi squared values
            for gpu_vis, gpu_X2 in S.run(gpu_ops):
                self.assertTrue(np.allclose(cpu_vis, gpu_vis))
                self.assertTrue(np.allclose(cpu_X2, gpu_X2))

if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import numpy as np

import montblanc

try:
    import pyrap.measures

    pm = pyrap.measures.measures()
except ImportError as e:
    pm = None
    montblanc.log.warning("python-casacore import failed. "
                       "Parallactic Angle computation will fail.")

def parallactic_angles(times, antenna_positions, field_centre, offsets=None):
    """
    Computes parallactic angles per timestep for the given
    reference antenna position and field centre.

    Arguments:
        times: ndarray
            Array of unique times with shape (ntime,),
            obtained from TIME column of MS table
        antenna_positions: ndarray of shape (na, 3)
            Antenna positions, obtained from POSITION
            column of MS ANTENNA sub-table
        field_centre : ndarray of shape (2,)
            Field centre, should be obtained from MS PHASE_DIR
        offsets: ndarray of shape (ntime, na, 2) or None
            Containing time-variable offsets in ra, dec from
            pointing centre per antenna. If None is passed offsets
            of zero are assumed.

    Returns:
        An array of parallactic angles per antenna per time-step

    """
    import pyrap.quanta as pq

    try:
        # Create direction measure for the zenith
        zenith = pm.direction('AZEL','0deg','90deg')
    except AttributeError as e:
        if pm is None:
            raise ImportError("python-casacore import failed")

        raise

    # Create position measures for each antenna
    reference_positions = [pm.position('itrf',
        *(pq.quantity(x,'m') for x in pos))
        for pos in antenna_positions]

    # Compute pointing centre in radians
    na = antenna_positions.shape[0]
    nt = times.shape[0]
    if offsets is None:
        offsets = np.zeros((nt, na, 2))

    fc_rad = np.asarray([[pm.direction('J2000',
                            pq.quantity(field_centre[0] + offsets[t, a, 0], 'rad'),
                            pq.quantity(field_centre[1] + offsets[t, a, 1], 'rad'))
                          for a in range(na)]
                         for t in range(nt)])

    return np.asarray([
            # Set current time as the reference frame
            pm.do_frame(pm.epoch("UTC", pq.quantity(t, "s")))
            and
            [   # Set antenna position as the reference frame
                pm.do_frame(rp)
                and
                pm.posangle(fc_rad[ti, ai], zenith).get_value("rad")
                for ai, rp in enumerate(reference_positions)
            ]
        for ti, t in enumerate(times)])



from os.path import exists as pexists
from os.path import join as pjoin

from astropy.io import fits
import numpy as np

from montblanc.impl.rime.tensorflow.sources.fits_beam_source_provider import (
                            REIM, LINEAR_CORRELATIONS, CIRCULAR_CORRELATIONS,
                            FitsFilenameTemplate, _create_filenames)

BITPIX_MAP = {
    np.int8: 8,
    np.uint8: 8,
    np.int16: 16,
    np.uint16: 16,
    np.int32: 32,
    np.uint32: 32,
    np.float32: -32,
    np.float64: -64,
}

DEFAULT_SCHEMA = pjoin("test_beam_$(corr)_$(reim).fits")

def beam_factory(polarisation_type='linear',
                    frequency=None,
                    dtype=np.float64,
                    schema=DEFAULT_SCHEMA,
                    overwrite=False):
    """ Generate a MeqTrees compliant beam cube """

    # MeerKAT l-band, 64 channels
    if frequency is None:
        frequency = np.linspace(.856e9, .856e9*2, 64,
                                endpoint=True, dtype=np.float64)
    # Generate a linear space of grid frequencies
    gfrequency = np.linspace(frequency[0], frequency[-1],
                            32, dtype=np.float64)
    bandwidth = gfrequency[-1] - frequency[0]
    bandwidth_delta = bandwidth / gfrequency.shape[0]-1

    if polarisation_type == 'linear':
        CORR = LINEAR_CORRELATIONS
    elif polarisation_type == 'circular':
        CORR = CIRCULAR_CORRELATIONS
    else:
        raise ValueError("Invalid polarisation_type %s" % polarisation_type)

    # List of key values of the form:
    #
    #    (key, None)
    #    (key, (value,))
    #    (key, (value, comment))
    #
    # We put them in a list so that they are added to the
    # FITS header in the correct order
    axis1 = [
        ("CTYPE", ('X', "points right on the sky")),
        ("CUNIT", ('DEG', 'degrees')),
        ("NAXIS", (257, "number of X")),
        ("CRPIX", (129, "reference pixel (one relative)")),
        ("CRVAL", (0.0110828777007, "degrees")),
        ("CDELT", (0.011082, "degrees"))]

    axis2 = [
        ("CTYPE", ('Y', "points up on the sky")),
        ("CUNIT", ('DEG', 'degrees')),
        ("NAXIS", (257, "number of Y")),
        ("CRPIX", (129, "reference pixel (one relative)")),
        ("CRVAL", (-2.14349358381E-07, "degrees")),
        ("CDELT", (0.011082, "degrees"))]

    axis3 = [
        ("CTYPE", ('FREQ', )),
        ("CUNIT", None),
        ("NAXIS", (gfrequency.shape[0], "number of FREQ")),
        ("CRPIX", (1, "reference frequency position")),
        ("CRVAL", (gfrequency[0], "reference frequency")),
        ("CDELT", (bandwidth_delta, "frequency step in Hz"))]

    axis4 = [
        ("CTYPE", ('STOKES', )),
        ("CUNIT", None),
        ("NAXIS", (1, "number of STOKES")),
        ("CRPIX", (1, "reference stokes index")),
        ("CRVAL", (1, "first stokes index")),
        ("CDELT", (-5,))]

    axes = [axis1, axis2, axis3]

    metadata = [
        ('SIMPLE', True),
        ('BITPIX', BITPIX_MAP[dtype]),
        ('NAXIS', len(axes)),
        ('OBSERVER', "Astronomer McAstronomerFace"),
        ('ORIGIN', "Artificial"),
        ('TELESCOP', "Telescope"),
        ('OBJECT', 'beam'),
        ('EQUINOX', 2000.0),
    ]

    # Create header and set metadata
    header = fits.Header()
    header.update(metadata)

    # Now set the key value entries for each axis
    ax_info = [('%s%d' % (k,a),) + vt
        for a, axis_data in enumerate(axes, 1)
        for k, vt in axis_data
        if vt is not None]
    header.update(ax_info)

    # Now setup the GFREQS
    # Jitter them randomly, except for the endpoints
    frequency_jitter = np.random.random(size=gfrequency.shape)-0.5
    frequency_jitter *= 0.1*bandwidth_delta
    frequency_jitter[0] = frequency_jitter[-1] = 0.0
    gfrequency += frequency_jitter

    # Check that gfrequency is monotically increasing
    assert np.all(np.diff(gfrequency) >= 0.0)

    for i, gfreq in enumerate(gfrequency, 1):
        header['GFREQ%d' % i] = gfreq

    # Figure out the beam filenames from the schema
    filenames = _create_filenames(schema, polarisation_type)

    for filename in [f for ri_pair in list(filenames.values()) for f in ri_pair]:
        if overwrite or not pexists(filename):
            ex = np.deg2rad(1.0)
            coords = np.linspace(-ex, ex, header['NAXIS2'], endpoint=True)

            r = np.sqrt(coords[None,:,None]**2 + coords[None,None,:]**2)
            fq = gfrequency[:,None,None]
            beam = np.cos(np.minimum(65*fq*1e-9*r, 1.0881))**3

            primary_hdu = fits.PrimaryHDU(beam, header=header)
            primary_hdu.writeto(filename, overwrite=overwrite)
        else:
            with fits.open(filename, mode='update', memmap=False) as file:
                uheader = file[0].header
                assert uheader['CTYPE3'] == 'FREQ'
                unchan = uheader['NAXIS3']

                bandwidth_delta = bandwidth / (unchan-1)
                uheader['CRVAL3'] = frequency[0]
                uheader['CDELT3'] = bandwidth_delta

                # Generate a linear space of grid frequencies
                # Jitter them randomly, except for the endpoints
                gfrequency = np.linspace(frequency[0], frequency[-1],
                                        unchan-1, dtype=np.float64)
                frequency_jitter = np.random.random(size=gfrequency.shape)-0.5
                frequency_jitter *= 0.1*bandwidth_delta
                frequency_jitter[0] = frequency_jitter[-1] = 0.0
                gfrequency += frequency_jitter

                # Check that gfrequency is monotically increasing
                assert np.all(np.diff(gfrequency) >= 0.0)

                # Remove existing GFREQ data
                try:
                    del uheader['GFREQ?*']
                except:
                    pass

                # Update existing GFREQ data
                for i, gfreq in enumerate(gfrequency, 1):
                    uheader['GFREQ%d' % i] = gfreq

    return filenames

if __name__ == "__main__":
    filenames = beam_factory(schema="test_beam_$(corr)_$(reim).fits")



import unittest

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tensorflow.python.client import device_lib


def radec_to_lm(radec, phase_centre):
        pc_ra, pc_dec = phase_centre
        sin_d0 = np.sin(pc_dec)
        cos_d0 = np.cos(pc_dec)

        da = radec[:, 0] - pc_ra
        sin_da = np.sin(da)
        cos_da = np.cos(da)

        sin_d = np.sin(radec[:, 1])
        cos_d = np.cos(radec[:, 1])

        lm = np.empty_like(radec)
        lm[:, 0] = cos_d*sin_da
        lm[:, 1] = sin_d*cos_d0 - cos_d*sin_d0*cos_da

        return lm


class TestRadecToLm(unittest.TestCase):
    """ Tests the RadecToLm operator """

    def setUp(self):
        # Load the custom operation library
        # Load the rime operation library
        from montblanc.impl.rime.tensorflow import load_tf_lib
        self.rime = load_tf_lib()
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                         if d.device_type == 'GPU']

    def test_radec_to_lm(self):
        """ Test the RadecToLm operator """
        # List of type constraint for testing this operator
        type_permutations = [np.float32, np.float64]

        # Run test with the type combinations above
        for FT in type_permutations:
            self._impl_test_radec_to_lm(FT)

    def _impl_test_radec_to_lm(self, FT):
        """ Implementation of the RadecToLm operator test """

        # Create input variables
        nsrc = 10
        radec = np.random.random(size=[nsrc, 2]).astype(FT)
        phase_centre = np.random.random(size=[2]).astype(FT)

        # Argument list
        np_args = [radec, phase_centre]
        # Argument string name list
        arg_names = ['radec', 'phase_centre']
        # Constructor tensorflow variables
        tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, arg_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return self.rime.radec_to_lm(*tf_args)

        # Pin operation to CPU
        cpu_op = _pin_op('/cpu:0', *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.compat.v1.global_variables_initializer()

        with tf.compat.v1.Session() as S:
            S.run(init_op)
            cpu_result = S.run(cpu_op)
            assert np.allclose(cpu_result, radec_to_lm(*np_args))

            for gpu_result in S.run(gpu_ops):
                assert np.allclose(cpu_result, gpu_result)

if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import tensorflow.compat.v1 as tf
tf.disable_eager_execution()
from tensorflow.python.client import device_lib


def rf(*s):
    return np.random.random(size=s)


class TestGaussShape(unittest.TestCase):
    """ Tests the GaussShape operator """
    def setUp(self):
        # Load the rime operation library
        from montblanc.impl.rime.tensorflow import load_tf_lib
        self.rime = load_tf_lib()
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                         if d.device_type == 'GPU']

    def test_gauss_shape(self):
        for FT in [np.float32, np.float64]:
            self._impl_test_gauss_shape(FT)

    def _impl_test_gauss_shape(self, FT):
        ngsrc, ntime, na, nchan = 10, 15, 7, 16
        nbl = na*(na-1)//2

        np_uvw = rf(ntime, na, 3).astype(FT)
        np_ant1, np_ant2 = [np.int32(x) for x in np.triu_indices(na, 1)]
        np_ant1, np_ant2 = (np.tile(np_ant1, ntime).reshape(ntime, nbl),
                            np.tile(np_ant2, ntime).reshape(ntime, nbl))
        np_frequency = np.linspace(1.4e9, 1.5e9, nchan).astype(FT)
        np_gauss_params = rf(3, ngsrc)*np.array([0.1, 0.1, 1.0])[:, np.newaxis]
        np_gauss_params = np_gauss_params.astype(FT)

        assert np_ant1.shape == (ntime, nbl), np_ant1.shape
        assert np_ant2.shape == (ntime, nbl), np_ant2.shape
        assert np_frequency.shape == (nchan,)

        np_args = [np_uvw, np_ant1, np_ant2, np_frequency, np_gauss_params]
        arg_names = ["uvw", "ant1", "ant2", "frequency", "gauss_params"]

        tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, arg_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return self.rime.gauss_shape(*tf_args)

        # Pin operation to CPU
        cpu_op = _pin_op('/cpu:0', *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.compat.v1.global_variables_initializer()

        with tf.compat.v1.Session() as S:
            S.run(init_op)

            cpu_gauss_shape = S.run(cpu_op)

            for gpu_gauss_shape in S.run(gpu_ops):
                self.assertTrue(np.allclose(cpu_gauss_shape,
                                            gpu_gauss_shape))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import collections
import functools
import unittest

from montblanc.impl.rime.tensorflow.sources.source_provider import (
    SourceProvider,
    find_sources,
    DEFAULT_ARGSPEC)

def constant_cache(method):
    """
    Caches constant arrays associated with an array name.

    The intent of this decorator is to avoid the cost
    of recreating and storing many arrays of constant data,
    especially data created by np.zeros or np.ones.
    Instead, a single array of the first given shape is created
    and any further requests for constant data of the same
    (or smaller) shape are served from the cache.

    Requests for larger shapes or different types are regarded
    as a cache miss and will result in replacement of the
    existing cache value.
    """
    @functools.wraps(method)
    def wrapper(self, context):
        # Defer to method if no caching is enabled
        if not self._is_cached:
            return method(self, context)

        name = context.name
        cached = self._constant_cache.get(name, None)

        # No cached value, call method and return
        if cached is None:
            data = self._constant_cache[name] = method(self, context)
            return data

        # Can we just slice the existing cache entry?
        # 1. Are all context.shape's entries less than or equal
        #    to the shape of the cached data?
        # 2. Do they have the same dtype?
        cached_ok = (cached.dtype == context.dtype and
            all(l <= r for l,r in zip(context.shape, cached.shape)))

        # Need to return something bigger or a different type
        if not cached_ok:
            data = self._constant_cache[name] = method(self, context)
            return data

        # Otherwise slice the cached data
        return cached[tuple(slice(0, s) for s in context.shape)]

    f = wrapper
    f.__decorator__ = constant_cache.__name__

    return f

def chunk_cache(method):
    """
    Caches chunks of default data.

    This decorator caches generated default data so as to
    avoid recomputing it on a subsequent queries to the
    provider.
    """

    @functools.wraps(method)
    def wrapper(self, context):
        # Defer to the method if no caching is enabled
        if not self._is_cached:
            return method(self, context)

        # Construct the key for the given index
        name = context.name
        idx = context.array_extents(name)
        key = tuple(i for t in idx for i in t)
        # Access the sub-cache for this array
        array_cache = self._chunk_cache[name]

        # Cache miss, call the function
        if key not in array_cache:
            array_cache[key] = method(self, context)

        return array_cache[key]

    f = wrapper
    f.__decorator__ = chunk_cache.__name__
    return f

class DefaultsSourceProvider(SourceProvider):
    def __init__(self, cache=False):
        self._is_cached = cache
        self._constant_cache = {}
        self._chunk_cache = collections.defaultdict(dict)

    def name(self):
        return self.__class__.__name__

    def clear_cache(self):
        self._constant_cache.clear()
        self._chunk_cache.clear()

class TestDefaultsSourceProvider(unittest.TestCase):

    def test_defaults_source_provider(self):
        import numpy as np
        import types

        # Create source provider and graft a model_vis method onto it
        defprov = DefaultsSourceProvider(cache=True)

        model_vis = lambda self, context: np.zeros(context.shape,
            context.dtype)

        defprov.model_vis = types.MethodType(constant_cache(model_vis),
            defprov)

        # Mock a context context object
        class Context(object):
            pass

        context = Context()
        context.name = 'model_vis'
        context.shape = (10, 16)
        context.dtype = np.float64

        A = defprov.model_vis(context)
        self.assertTrue(A.flags['OWNDATA'])
        self.assertEqual(A.shape, context.shape)

        context.name = 'model_vis'
        context.shape = supplied_shape = (100, 32)
        context.dtype = np.float64

        B = defprov.model_vis(context)
        self.assertTrue(B.flags['OWNDATA'])
        self.assertEqual(B.shape, context.shape)

        context.name = 'model_vis'
        context.shape = (8, 2)
        context.dtype = np.float64

        C = defprov.model_vis(context)
        self.assertFalse(C.flags['OWNDATA'])
        self.assertEqual(C.shape, context.shape)
        self.assertIs(C.base, B)

        cached_shape = defprov._constant_cache['model_vis'].shape
        self.assertEqual(cached_shape, supplied_shape)

if __name__ == "__main__":
    unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import logging
import unittest
import numpy as np
import tempfile
import time

import montblanc
import montblanc.factory
import montblanc.util as mbu

class TestUtils(unittest.TestCase):
    """
    TestUtil class defining unit tests for
    montblanc's montblanc.util module
    """

    def setUp(self):
        """ Set up each test case """
        np.random.seed(int(time.time()) & 0xFFFFFFFF)
        montblanc.setup_test_logging()

    def tearDown(self):
        """ Tear down each test case """
        pass

    def test_baseline_antenna_nrs(self):
        """ Test conversion between antenna and baseline numbers """

        def do_check(na, nbl, nbl_auto):
            # get nr baselines from nr of antenna
            self.assertTrue(mbu.nr_of_baselines(na) == nbl)
            self.assertTrue(mbu.nr_of_baselines(na, False) == nbl)
            self.assertTrue(mbu.nr_of_baselines(na, True) == nbl_auto)

            # get nr antenna from nr of baselines
            self.assertTrue(mbu.nr_of_antenna(nbl) == na)
            self.assertTrue(mbu.nr_of_antenna(nbl, False) == na)
            self.assertTrue(mbu.nr_of_antenna(nbl_auto, True) == na)

        do_check(7, 7*6//2, 7*8//2)                  # KAT7
        do_check(14, 14*13//2, 14*15//2)             # Westerbork
        do_check(27, 27*26//2, 27*28//2)             # VLA
        do_check(64, 64*63//2, 64*65//2)             # MeerKAT
        do_check(3500, 3500*3499//2, 3500*3501//2)   # SKA

    def test_random_like(self):
        """
        Test that the random_like function produces sensible data
        """

        # Try for floats and complex data
        for dtype in [np.float32, np.float64, np.complex64, np.complex128]:
            # Test random array creation with same
            # shape and type as existing array
            shape = (np.random.randint(1, 50), np.random.randint(1, 50))
            ary = np.empty(shape=shape, dtype=dtype)    
            random_ary = mbu.random_like(ary)

            # Test that that the shape and type is correct
            self.assertTrue(random_ary.shape == ary.shape)
            self.assertTrue(random_ary.dtype == dtype)

            # Test that we're getting complex data out
            if np.issubdtype(dtype, np.complexfloating):
                proportion_cplx = np.sum(np.iscomplex(random_ary)) / random_ary.size
                self.assertTrue(proportion_cplx > 0.9)

            # Test random array creation with supplied shape and type
            shape = (np.random.randint(1, 50), np.random.randint(1, 50))
            random_ary = mbu.random_like(shape=shape, dtype=dtype)

            # Test that that the shape and type is correct
            self.assertTrue(random_ary.shape == shape)
            self.assertTrue(random_ary.dtype == dtype)

            # Test that we're getting complex data out
            if np.issubdtype(dtype, np.complexfloating):
                proportion_cplx = np.sum(np.iscomplex(random_ary)) / random_ary.size
                self.assertTrue(proportion_cplx > 0.9)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestUtils)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from ..context_help import context_help
from ..hypercube_proxy_metaclass import HypercubeProxyMetaClass
import six

class _setter_property(object):
    def __init__(self, func, doc=None):
        self.func = func
        self.__doc__ = doc if doc is not None else func.__doc__

    def __set__(self, obj, value):
        return self.func(obj, value)

@six.add_metaclass(HypercubeProxyMetaClass)
class SinkContext(object):
    """
    Context object passed to data sinks.

    Primarily, it exists to provide a tile of output data to the user.

    .. code-block:: python

        class MySinkProvider(SinkProvider):
            vis_queue = Queue(10)

            ...
            def model_vis(self, context):
                print context.help(display_cube=True)
                # Consume data
                vis_queue.put(context.data)


    Public methods of a :py:class:`~hypercube.base_cube.HyperCube`
    are proxied on this object. Other useful information, such
    as the configuration, iteration space arguments and the
    abstract array schema are also present on this object.
    """

    __slots__ = ('_cube', '_cfg', '_name', '_data', '_input_cache',
        '_cube_attributes', '_iter_args', '_array_schema')

    def __init__(self, name, cube, slvr_cfg,
            iter_args, array_schema,
            data, input_cache):

        self._name = name
        self._cube = cube
        self._iter_args = iter_args
        self._array_schema = array_schema
        self._cfg = slvr_cfg
        self._data = data
        self._input_cache = input_cache

    @_setter_property
    def cube(self, value):
        self._cube = value

    @property
    def cfg(self):
        """ Configuration """
        return self._cfg

    @cfg.setter
    def cfg(self, value):
        self._cfg = value

    @property
    def iter_args(self):
        """
        Iteration arguments that describe the tile sizes
        over which iteration is performed. In the following example,
        iteration is occuring in tiles of 100 Timesteps, 64 Channels
        and 50 Point Sources.

        .. code-block:: python

            context.iter_args == [("ntime", 100),
                    ("nchan", 64), ("npsrc", 50)]
        """
        return self._iter_args

    @property
    def array_schema(self):
        """
        The array schema of the array associated
        with this data source. For instance if `model_vis` is
        registered on a hypercube as follows:

        .. code-block:: python

            # Register model_vis array_schema on hypercube
            cube.register_array("model_vis",
                ("ntime", "nbl", "nchan", "ncorr"),
                np.complex128)

            ...
            # Create a source context for model_vis data source
            context = SourceContext("model_vis", ...)
            ...
            # Obtain the array schema
            context.array_schema == ("ntime", "nbl", "nchan", "ncorr")

        """
        return self._array_schema

    @property
    def data(self):
        """
        The data tile available for consumption by the associated sink
        """
        return self._data

    @property
    def input(self):
        """
        The dictionary of inputs used to produce
        :py:obj:`~SinkContext.data`. For example, if one
        wished to find the antenna pair used to produce a
        particular model visibility, one could do the following:

        .. code-block:: python

            def model_vis(self, context):
                ant1 = context.input["antenna1"]
                ant2 = context.input["antenna2"]
                model_vis = context.data

        """
        return self._input_cache

    @property
    def name(self):
        """ The name of the data sink of this context. """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    def help(self, display_cube=False):
        """
        Get help associated with this context

        Parameters
        ----------
        display_cube: bool
            Add hypercube description to the output
        Returns
        -------
            str
                A help string associated with this context
        """
        return context_help(self, display_cube)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import inspect
import os

from montblanc.logsetup import setup_logging, setup_test_logging
from montblanc.tests import test
import pkg_resources
__version__ = pkg_resources.get_distribution("montblanc").version

log = setup_logging()

# This solution for constants based on
# http://stackoverflow.com/a/2688086
# Create a property that throws when
# you try and set it
def constant(f):
    def fset(self, value):
        raise SyntaxError('Foolish Mortal! You would dare change a universal constant?')
    def fget(self):
        return f()

    return property(fget, fset)

class MontblancConstants(object):
    # The speed of light, in metres
    @constant
    def C():
        return 299792458

# Create a constants object
constants = MontblancConstants()

def rime_solver_cfg(**kwargs):
    """
    Produces a SolverConfiguration object, inherited from
    a simple python dict, and containing the options required
    to configure the RIME Solver.

    Keyword arguments
    -----------------
    Any keyword arguments are inserted into the
    returned dict.

    Returns
    -------
    A SolverConfiguration object.
    """
    from .configuration import (load_config, config_validator,
        raise_validator_errors)

    def _merge_copy(d1, d2):
        return { k: _merge_copy(d1[k], d2[k]) if k in d1
                                                and isinstance(d1[k], dict)
                                                and isinstance(d2[k], dict)
                                            else d2[k] for k in d2 }

    try:
        cfg_file = kwargs.pop('cfg_file')
    except KeyError as e:
        slvr_cfg = kwargs
    else:
        cfg = load_config(cfg_file)
        slvr_cfg = _merge_copy(cfg, kwargs)

    # Validate the configuration, raising any errors
    validator = config_validator()
    validator.validate(slvr_cfg)
    raise_validator_errors(validator)

    return validator.document

def rime_solver(slvr_cfg):
    """
    rime_solver(slvr_cfg)

    Returns a solver suitable for solving the RIME.

    Parameters
    ----------
    slvr_cfg : RimeSolverConfiguration
            Solver Configuration.

    Returns
    -------
    A solver
    """

    import montblanc.factory

    return montblanc.factory.rime_solver(slvr_cfg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import logging
import numpy as np
import os

import montblanc
import montblanc.util as mbu

from montblanc.impl.rime.tensorflow.ms import MeasurementSetManager
from montblanc.impl.rime.tensorflow.sources import (SourceProvider,
    FitsBeamSourceProvider,
    MSSourceProvider)
from montblanc.impl.rime.tensorflow.sinks import (SinkProvider,
    MSSinkProvider)

class RadioSourceProvider(SourceProvider):
    """ Simulates a number of equally defined point sources """

    def __init__(self, npsrc):
        """ Simulate npsrc point sources """
        self._npsrc = npsrc

    def name(self):
        return "TF example"

    def point_lm(self, context):
        """ Return a lm coordinate array to montblanc """
        lm = np.empty(context.shape, context.dtype)

        # Print the array schema
        montblanc.log.info(context.array_schema.shape)
        # Print the space of iteration
        montblanc.log.info(context.iter_args)

        (ls, us) = context.dim_extents('npsrc')

        lm[:,0] = 0.0008
        lm[:,1] = 0.0036

        lm[:,:] = 0
        return lm

    def point_stokes(self, context):
        """ Return a stokes parameter array to montblanc """
        stokes = np.empty(context.shape, context.dtype)
        stokes[:,:,0] = 1
        stokes[:,:,1:4] = 0
        return stokes

    def point_alpha(self, context):
        """ Return a spectral index (alpha) array to montblanc """
        return np.zeros(context.shape, context.dtype)

    def frequency(self, context):
        """ Return a frequency array to montblanc """
        return np.full(context.shape, 1.415e9, context.dtype)

    def ref_frequency(self, context):
        """ Return a reference frequency array to montblanc """
        ref_freq = np.empty(context.shape, context.dtype)
        ref_freq[:] = 1.415e9

        return ref_freq

    def updated_dimensions(self):
        """ Tell montblanc about dimension sizes (point sources only) """
        return [('npsrc', self._npsrc)]

class RimeSinkProvider(SinkProvider):
    def name(self):
        return 'Sink'

    def model_vis(self, context):
        montblanc.log.info(context.data.ravel()[0:128].reshape(-1,4))
        montblanc.log.info(context.data.mean())
        montblanc.log.info(context.data.sum())

if __name__ == '__main__':
    import sys
    import argparse
    import tensorflow as tf

    parser = argparse.ArgumentParser(description='RIME MS test script')
    parser.add_argument('msfile', help='Measurement Set File')
    parser.add_argument('-b', '--beam',
        type=str, default='', help='Base beam filename')
    parser.add_argument('-np','--npsrc',dest='npsrc',
        type=int, default=10, help='Number of Point Sources')
    parser.add_argument('-ac','--auto-correlations',dest='auto_correlations',
        type=lambda v: v.lower() in ("yes", "true", "t", "1"),
        choices=[True, False], default=False,
        help='Handle auto-correlations')

    args = parser.parse_args(sys.argv[1:])

    # Set the logging level
    montblanc.log.setLevel(logging.DEBUG)
    [h.setLevel(logging.DEBUG) for h in montblanc.log.handlers]

    slvr_cfg = montblanc.rime_solver_cfg(
        mem_budget=1024*1024*1024,
        data_source='default',
        dtype='double',
        auto_correlations=args.auto_correlations)

    with montblanc.rime_solver(slvr_cfg) as slvr:
        # Manages measurement sets
        ms_mgr = MeasurementSetManager(args.msfile, slvr_cfg)

        source_provs = []
        # Read problem info from the MS, taking observed visibilities from MODEL_DAT
        source_provs.append(MSSourceProvider(ms_mgr, 'MODEL_DATA'))
        # Add a beam when you're ready
        #source_provs.append(FitsBeamSourceProvider('beam_$(corr)_$(reim).fits'))
        source_provs.append(RadioSourceProvider(args.npsrc))

        sink_provs = []
        # Dump model visibilities into CORRECTED_DATA
        sink_provs.append(MSSinkProvider(ms_mgr, 'CORRECTED_DATA'))

        slvr.solve(source_providers=source_provs,
            sink_providers=sink_provs)



import attridict as AttrDict

import tensorflow as tf
from tensorflow.python.ops import data_flow_ops

from .queue_wrapper import _get_queue_types

class StagingAreaWrapper(object):
    def __init__(self, name, fed_arrays, data_sources, shared_name=None):
        self._name = name
        self._fed_arrays = fed_arrays
        self._data_sources = data_sources

        # Infer types of the given fed_arrays
        self._dtypes = _get_queue_types(fed_arrays, data_sources)

        # Create placeholders for the fed arrays
        self._placeholders = placeholders = [tf.compat.v1.placeholder(dt,
                name="{n}_placeholder".format(n=n))
            for n, dt in zip(fed_arrays, self._dtypes)]

        self._staging_area = sa = data_flow_ops.StagingArea(self._dtypes,
            names=fed_arrays, shared_name=shared_name)

        self._put_op = sa.put({n: p for n, p in zip(fed_arrays, placeholders)})
        self._get_op = sa.get()

    @property
    def staging_area(self):
        return self._staging_area

    @property
    def fed_arrays(self):
        return self._fed_arrays

    @property
    def placeholders(self):
        return self._placeholders

    def put(self, data):
        return self._staging_area.put(data)

    def put_from_list(self, data):
        return self.put({n: d for n,d in zip(self._fed_arrays, data)})

    def get(self):
        return self._staging_area.get()

    def get_to_list(self):
        D = self.get()
        return [D[n] for n in self._fed_arrays]

    def get_to_attrdict(self):
        return AttrDict(**self.get())

    @property
    def put_op(self):
        return self._put_op

    @property
    def get_op(self):
        return self._get_op


def create_staging_area_wrapper(name, fed_arrays, data_source, *args, **kwargs):
    return StagingAreaWrapper(name, fed_arrays, data_source, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import inspect

class AbstractSinkProvider(object):

    def name(self):
        """ Returns this data sink's name """
        raise NotImplementedError()

    def init(self, init_context):
        """ Called when initialising Providers """
        raise NotImplementedError()

    def start(self, start_context):
        """ Called at the start of any solution """
        raise NotImplementedError()

    def stop(self, stop_context):
        """ Called at the end of any solution """
        raise NotImplementedError()

    def close(self):
        """ Perform any required cleanup """
        raise NotImplementedError()

    def clear_cache(self):
        """ Clears any caches associated with the sink """
        raise NotImplementedError()

    def sinks(self):
        """ Returns a dictionary of sink methods, keyed on sink name """
        raise NotImplementedError()

def find_sinks(obj):
    """
    Returns a dictionary of sink methods found on this object,
    keyed on method name. Sink methods are identified by
    (self, context) arguments on this object. For example:

    def f(self, context):
        ...

    is a sink method, but

    def f(self, ctx):
        ...

    is not.

    """
    SINK_ARGSPEC = ['self', 'context']

    return { n: m for n, m in inspect.getmembers(obj, inspect.ismethod)
        if inspect.getfullargspec(m)[0] == SINK_ARGSPEC }

class SinkProvider(AbstractSinkProvider):

    def init(self, init_context):
        """ Called when initialising Providers """
        pass

    def start(self, start_context):
        """ Called at the start of any solution """
        pass

    def stop(self, stop_context):
        """ Called at the end of any solution """
        pass

    def close(self):
        """ Perform any required cleanup. NOOP """
        pass

    def clear_cache(self):
        """ Clears any caches associated with the sink """
        pass

    def sinks(self):
        """
        Returns a dictionary of sink methods found on this object,
        keyed on method name. Sink methods are identified by
        (self, context) arguments on this object. For example:

        def f(self, context):
            ...

        is a sink method, but

        def f(self, ctx):
            ...

        is not.

        """

        try:
            return self._sinks
        except AttributeError:
            self._sinks = find_sinks(self)

        return self._sinks

    def __str__(self):
        return self.name()




import argparse

import numpy as np

import montblanc
from montblanc.impl.rime.tensorflow.sources import SourceProvider
from montblanc.impl.rime.tensorflow.sinks import SinkProvider

def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--ntime", default=100, type=int,
                                   help="Number of timesteps")
    parser.add_argument("--nchan", default=64, type=int,
                                   help="Number of channels")
    parser.add_argument("--na", default=27, type=int,
                                    help="Number of antenna")

    return parser

args = create_parser().parse_args()

# Two point sources at centre

# LM coordinates
lm_coords = [(0.0, 0.0),
             (0.0, 0.0)]
# Stokes parameters (I, Q, U, V)
lm_stokes = [(1.0, 0.0, 0.0, 0.0),
             (1.0, 0.0, 0.0, 0.0)]

class CustomSourceProvider(SourceProvider):
    """
    Supplies data to montblanc via data source methods,
    which have the following signature.

    .. code-block:: python

        def point_lm(self, context)
            ...
    """
    def name(self):
        """ Name of Source Provider """
        return self.__class__.__name__

    def updated_dimensions(self):
        """ Inform montblanc about dimension sizes """
        return [("ntime", args.ntime),      # Timesteps
                ("nchan", args.nchan),      # Channels
                ("na", args.na),            # Antenna
                ("npsrc", len(lm_coords))]  # Number of point sources

    def point_lm(self, context):
        """ Supply point source lm coordinates to montblanc """

        # Shape (npsrc, 2)
        (ls, us), _ = context.array_extents(context.name)
        return np.asarray(lm_coords[ls:us], dtype=context.dtype)

    def point_stokes(self, context):
        """ Supply point source stokes parameters to montblanc """

        # Shape (npsrc, ntime, nchan, 4)
        (ls, us), (lt, ut), (lc, uc), (l, u) = context.array_extents(context.name)

        data = np.empty(context.shape, context.dtype)
        data[ls:us,:,:,l:u] = np.asarray(lm_stokes)[ls:us,None,None,:]
        return data

    def uvw(self, context):
        """ Supply UVW antenna coordinates to montblanc """

        # Shape (ntime, na, 3)
        (lt, ut), (la, ua), (l, u) = context.array_extents(context.name)

        # Create empty UVW coordinates
        data = np.empty(context.shape, context.dtype)
        data[:,:,0] = np.arange(la+1, ua+1)    # U = antenna index
        data[:,:,1] = 0                        # V = 0
        data[:,:,2] = 0                        # W = 0

        return data

class CustomSinkProvider(SinkProvider):
    """
    Receives data from montblanc via data sink methods,
    which have the following signature

    .. code-block:: python

        def model_vis(self, context):
            print context. data
    """
    def name(self):
        """ Name of the Sink Provider """
        return self.__class__.__name__

    def model_vis(self, context):
        """ Receive model visibilities from Montblanc in `context.data` """
        print((context.data))

# Configure montblanc solver with a memory budget of 2GB
# and set it to double precision floating point accuracy
slvr_cfg = montblanc.rime_solver_cfg(
    mem_budget=2*1024*1024*1024,
    dtype='double')

# Create montblanc solver
with montblanc.rime_solver(slvr_cfg) as slvr:
    # Create Customer Source and Sink Providers
    source_provs = [CustomSourceProvider()]
    sink_provs = [CustomSinkProvider()]

    # Call solver, supplying source and sink providers
    slvr.solve(source_providers=source_provs,
            sink_providers=sink_provs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Simon Perkins
#
# This file is part of montblanc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import logging
import logging.handlers

def setup_logging():
    """ Setup logging configuration """

    # Console formatter, mention name
    cfmt = logging.Formatter(('%(name)s - %(levelname)s - %(message)s'))

    # File formatter, mention time
    ffmt = logging.Formatter(('%(asctime)s - %(levelname)s - %(message)s'))

    # Console handler
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    ch.setFormatter(cfmt)

    # File handler
    fh = logging.handlers.RotatingFileHandler('montblanc.log',
        maxBytes=10*1024*1024, backupCount=10)
    fh.setLevel(logging.INFO)
    fh.setFormatter(ffmt)

    # Create the logger,
    # adding the console and file handler
    mb_logger = logging.getLogger('montblanc')
    mb_logger.handlers = []
    mb_logger.setLevel(logging.DEBUG)
    mb_logger.addHandler(ch)
    mb_logger.addHandler(fh)

    # Set up the concurrent.futures logger
    cf_logger = logging.getLogger('concurrent.futures')
    cf_logger.setLevel(logging.DEBUG)
    cf_logger.addHandler(ch)
    cf_logger.addHandler(fh)

    return mb_logger

def setup_test_logging():
    # Console formatter, mention name
    cfmt = logging.Formatter(('%(name)s - %(levelname)s - %(message)s'))

    # File formatter, mention time
    ffmt = logging.Formatter(('%(asctime)s - %(levelname)s - %(message)s'))

    # Only warnings and more serious stuff on the console
    ch = logging.StreamHandler()
    ch.setLevel(logging.WARN)
    ch.setFormatter(cfmt)

    # Outputs DEBUG level logging to file
    fh = logging.FileHandler('test.log')
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(ffmt)

    # Set up the montblanc logger
    mb_logger = logging.getLogger('montblanc')
    mb_logger.handlers = []
    mb_logger.setLevel(logging.DEBUG)
    mb_logger.addHandler(ch)
    mb_logger.addHandler(fh)

    # Set up the concurrent.futures logger
    cf_logger = logging.getLogger('concurrent.futures')
    cf_logger.setLevel(logging.DEBUG)
    cf_logger.addHandler(ch)
    cf_logger.addHandler(fh)

    return mb_logger